"""Model definition for Records return type."""

import logging
from enum import Enum

_logger = logging.getLogger(__name__)


class IncludeField(str, Enum):
    """Enumerates the valid types for a extra fields that can be returned with records."""

    FIELDS = "fields"
    METADATA_ID = "metadataId"

    def __str__(self) -> str:
        """Return the value as a string rather than a string of the full enum."""
        return self.value

    def build(self) -> str:
        """Return an IncludeField return model definition for QueryDocument."""
        return str(self)


class Records:
    """Records return model."""

    __match_args__ = ("include",)
    _include: list[IncludeField] | None

    def __init__(self, include: list[IncludeField | str] | None = None):
        """Initialize to base records.  Filter Fields and Metadata are optionally returned."""
        self._include = self._validate_include_fields(include)

    def __repr__(self) -> str:
        """Class representation."""
        return f"Records(include={self.include})"

    def __str__(self) -> str:
        """Class string formatting."""
        return f"Records(include={self.include})"

    @property
    def include(self) -> list[IncludeField] | None:
        """The include property."""
        return self._include

    def _validate_include_fields(self, include: list[IncludeField | str] | None) -> list[IncludeField] | None:
        match include:
            case None:
                return None
            case list() if len(include) > 0:
                include_fields = []
                for field in include:
                    match field:
                        case IncludeField():
                            include_fields.append(field)
                        case str():
                            include_fields.append(IncludeField(field))
                        case _:
                            raise ValueError(f"{field} is not a valid IncludeField")
                return include_fields
            case _:
                raise ValueError(f"{include} is not a valid list[IncludeField]")

    def build(self) -> dict[str, str | list[str]]:
        """Return a records return model definition for QueryDocument."""
        result: dict[str, str | list[str]] = {"type": "records"}
        if self.include:
            result.update({"include": [field.build() for field in self.include]})
        return result



"""Module for geometry models."""

# ruff: noqa: F401, I001 (unused-import, un-sorted)
# isort: skip_file
from typing import TypeAlias

# Point needs to be imported before BBox or Polygon because it's a dependency in both.
# Rexported here for easier access.
from .point import Point
from .bbox import BBox
from .polygon import Polygon, RawCoords



"""Polygon model definition."""

from typing import Any, TypeAlias

from typing_extensions import Self

from dfi.errors import LinearRingError, PolygonNotClosedError, PolygonUndefinedError
from dfi.models.filters.geometry import Point

RawCoords: TypeAlias = tuple[float, float]
"""Alias for `tuple[float, float]`"""

MIN_VERTICES = 4


class Polygon:
    """A 2D Polygon as defined in GeoJSON https://datatracker.ietf.org/doc/html/rfc7946#section-3.1.6.

    ??? attention
        Polygons should be composed of a singular linear ring.  MultiPolygons are not supported.

    """

    _coordinates: tuple[Point, ...]

    def __repr__(self) -> str:
        """Class representation."""
        return f"Polygon({self._coordinates})"

    def __str__(self) -> str:
        """Class string formatting."""
        return f"Polygon({self._coordinates})"

    @property
    def coordinates(self) -> tuple[Point, ...]:
        """The coordinates property."""
        return self._coordinates

    def from_geojson(self, geojson: dict[str, Any]) -> Self:
        """Create a Polygon from a given GeoJSON Polygon.

        Parameters
        ----------
        geojson:
            a geojson dictionary.

        Raises
        ------
        LinearRingError
        PolygonNotClosedError
        PolygonUndefinedError

        Examples
        --------
        ### Polygon from Points
        ```python
        geojson = {
            "type": "Polygon",
            "coordinates": [
                [
                    [0.0, 0.0],
                    [1.0, 0.0],
                    [1.0, 1.0],
                    [0.0, 1.0],
                    [0.0, 0.0],
                ]
            ],
        }
        Polygon().from_geojson(geojson)
        ```
        ```python
        Polygon(
            (
                Point(0.0, 0.0),
                Point(0.0, 1.0),
                Point(1.0, 1.0),
                Point(1.0, 0.0),
                Point(0.0, 0.0),
            )
        )
        ```
        """
        match geometry := geojson.get("type"):
            case "Polygon":
                coordinates = geojson["coordinates"][0]
                self._coordinates = tuple(Point(lon, lat) for lon, lat in coordinates)
            case "FeatureCollection" | "Feature" | "MultiPolygon" | "LineString" | "Point":
                raise TypeError(f"'geojson' is not a Polygon, found {geometry}.")
            case _:
                raise TypeError("'geojson' could not be parsed.")

        return self.validate()

    def from_points(self, coordinates: list[Point], geojson: bool = True) -> Self:
        """Create a Polygon from given coordinates.

        Parameters
        ----------
        coordinates:
            a list of Points.
        geojson:
            indicates if coordinate order follow GeoJSON specification.

            - if `True` expects coordinates with the form (longitude, latitude)
            - if `False` expects coordinates with the form (latitude, longitude)

        Returns
        -------
        Polygon

        Raises
        ------
        LinearRingError
        PolygonNotClosedError
        PolygonUndefinedError

        Examples
        --------
        ### Polygon from Points
        ```python
        raw_coords = [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 1.0],
            [0.0, 0.0],
        ]
        points = [Point(p.lon, p.lat) for p in raw_coords]
        Polygon().from_points(points)
        ```
        ```python
        Polygon(
            (
                Point(0.0, 0.0),
                Point(0.0, 1.0),
                Point(1.0, 1.0),
                Point(1.0, 0.0),
                Point(0.0, 0.0),
            )
        )
        ```
        """
        match coordinates:
            case list() | tuple():
                pass
            case _:
                raise TypeError("'coordinates' is not of type list[float].")
        if geojson:
            self._coordinates = tuple(coordinates)
        else:
            self._coordinates = tuple(Point(point.lat, point.lon) for point in coordinates)

        return self.validate()

    def from_raw_coords(self, coordinates: list[RawCoords], geojson: bool = True) -> Self:
        """Create a Polygon from given coordinates.

        Parameters
        ----------
        coordinates:
            a list of RawCoords (i.e. list[float, float]).
        geojson:
            indicates if coordinate order follow GeoJSON specification.

            - if `True` expects coordinates with the form (longitude, latitude)
            - if `False` expects coordinates with the form (latitude, longitude)

        Returns
        -------
        Polygon

        Raises
        ------
        LinearRingError
        PolygonNotClosedError
        PolygonUndefinedError

        Examples
        --------
        ### Polygon from RawCoords
        From RawCoords is useful when the coordinates of the Polygon are in a list.
        ```python
        raw_coords = [[0.0, 0.0], [1.0, 0.0], [1.0, 1.0], [0.0, 1.0], [0.0, 0.0]]
        points =
        Polygon().from_raw_coords(raw_coords)
        ```
        ```python
        Polygon(
            (
                Point(0.0, 0.0),
                Point(0.0, 1.0),
                Point(1.0, 1.0),
                Point(1.0, 0.0),
                Point(0.0, 0.0),
            )
        )
        ```
        """
        if geojson:
            self._coordinates = tuple(Point(lon, lat) for lon, lat in coordinates)
        else:
            self._coordinates = tuple(Point(lon, lat) for lat, lon in coordinates)

        return self.validate()

    def validate(self) -> Self:
        """Validate the Polygon.

        Returns
        -------
        Polygon

        Raises
        ------
        LinearRingError
        PolygonNotClosedError
        PolygonUndefinedError
        """
        if not hasattr(self, "_coordinates"):
            raise PolygonUndefinedError

        if len(self._coordinates) < MIN_VERTICES:
            raise LinearRingError(
                f"Polygons should be a linear rings with four or more points - only {len(self._coordinates)} found."
            )
        if self._coordinates[0] != self._coordinates[-1]:
            raise PolygonNotClosedError("Polygons should be a linear ring - first and last points are not identical.")

        return self

    def build(self) -> dict[str, str | tuple[RawCoords, ...]]:
        """Validate the polygon is defined and format the polygon for the Query Document."""
        self.validate()
        return {"type": "Polygon", "coordinates": tuple(point.build() for point in self._coordinates)}



"""Model definition for Count return type."""

import logging
from enum import Enum

_logger = logging.getLogger(__name__)


class GroupBy(str, Enum):
    """Enumerates the valid groupby fields."""

    UNIQUE_IDS = "uniqueId"

    def __str__(self) -> str:
        """Return the value as a string rather than a string of the full enum."""
        return self.value

    def build(self) -> dict[str, dict[str, str]]:
        """Return a GroupBy return model definition for QueryDocument."""
        return {"groupBy": {"type": str(self)}}


class Count:
    """Count return model."""

    __match_args__ = ("groupby",)
    _groupby: GroupBy | None

    def __init__(self, groupby: GroupBy | str | None = None):
        """Initialize to Coutn return model. Optionally specify a groupby field."""
        self._groupby = self._validate_groupby(groupby)

    def __repr__(self) -> str:
        """Class representation."""
        return f"Count(groupby={self._groupby})"

    def __str__(self) -> str:
        """Class string formatting."""
        return f"Count(groupby={self._groupby})"

    def _validate_groupby(self, groupby: GroupBy | str | None) -> GroupBy | None:
        """Validate the include list.

        Parameters
        ----------
        include:
            Can be a GroupBy object or a str that can be used to create a GroupBy object or None.
        """
        match groupby:
            case None | GroupBy():
                return groupby
            case str():
                return GroupBy(groupby)
            case _:
                raise ValueError(f"{groupby} is not a valid GroupBy")

    @property
    def groupby(self) -> GroupBy | None:
        """The groupby property."""
        return self._groupby

    def build(self) -> dict:
        """Return a count return model definition for QueryDocument."""
        result: dict[str, str | dict[str, str]] = {"type": "count"}
        if self._groupby:
            result.update(self._groupby.build())
        return result



"""Model definitions for filters."""

# ruff: noqa: F401 (unused-import)
from enum import Enum

from .filter_fields import FieldType, FilterField, FilterOperator
from .only import Only
from .time_range import TimeRange



"""Enumerated errors for possible DFI errors that may arise.

??? note
    Errors and Warnings prefixed with `DFI` (e.g. `DFI<name>Error`) indicates it 
    originates from within the GS Platform API.
"""


class DFIResponseError(Exception):
    """Raised when an error propagated back from the DFI API."""


class UnreachableError(Exception):
    """Indicates a section of code was reached that should be unreachable."""


class LinearRingError(Exception):
    """Raised when a Polygon is not a linear ring (<4 points)."""


class PolygonNotClosedError(Exception):
    """Raised when a Polygon not closed (first point is identical to last point)."""


class PolygonUndefinedError(Exception):
    """Raised when a Polygon is uninitialized."""


class LongitudeOutOfBoundsError(Exception):
    """Raised when a longitude value is not within (-180.0 < longitude < 180.0)."""


class LatitudeOutOfBoundsError(Exception):
    """Raised when a latitude value is not within (-90.0 < latitude < 90.0)."""


class AltitudeOutOfBoundsError(Exception):
    """Raised when an altitude value is not within (`-1.7976931348623157e308 < altitude < 1.7976931348623157e308`)."""


class BBoxLongitudeMismatchError(Exception):
    """Raised when `min_lon >= max_lon`."""


class BBoxLatitudeMismatchError(Exception):
    """Raised when `min_lat >= max_lat`."""


class BBoxValueError(Exception):
    """Raised when more than 4 values are used to create a BBox."""


class BBoxUndefinedError(Exception):
    """Raised when a BBox is uninitialized."""


class TimeRangeOutOfBoundsError(Exception):
    """Raised when a TimeRange value is not within.

    - datetime form:          (`1970-01-01 00:00:00+00:00 <= datetime <= 2038-01-19 03:14:08+00:00`)
    - unix epoch millis form: (`0 <= datetime <= 2,147,483,647`)
    """


class TimeRangeMismatchError(Exception):
    """Raised when `min_time > max_time`."""


class TimeRangeUndefinedError(Exception):
    """Raised when a TimeRange is uninitialized."""


class TimeZoneUndefinedError(Exception):
    """Raised when a TimeRange is defined without specifying a timezone."""


class FilterFieldNameNotInSchema(Exception):
    """Raised when a Filter Field is given a name that is not in the dataset schema."""


class FilterFieldValueError(Exception):
    """Raised when a Filter Field value is not a valid value for the field type."""


class FilterFieldTypeError(Exception):
    """Raised when a Filter Field value is not a valid value for the field type."""


class FilterFieldOperationValueError(Exception):
    """Raised when a Filter Field operation is not a valid value for the field type."""


class FilterFieldInvalidNullability(Exception):
    """Raised when a Filter Field nullability does not match the schema definition."""


class UnkownMessageReceivedError(Exception):
    """Raised when an unknown Server Side Event message is received."""


class NoFinishMessageReceivedError(Exception):
    """Raised when a stream ends without receving a 'finish' event."""


class EventsMissedError(Exception):
    """Raised when a fewer events are received than were sent."""


class NoEventsRecievedError(Exception):
    """Raised when 0 events received from the DFI API."""


class InvalidQueryDocument(Exception):
    """Raised when a QueryDocument is invalid."""



"""Manages identity authentication via the Identity Management API service."""

from typing import Any

from dfi.connect import Connect


class Identities:
    """Class responsible handling identities and tokens."""

    def __init__(self, conn: Connect) -> None:
        self.conn = conn

    def __repr__(self) -> str:
        """Class representation."""
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        """Class string formatting."""
        return f"""Instance of dfi.{self.__class__.__name__} composed with: {self.conn!s}"""

    def get_tokens(self, include_expired: bool = False) -> list[dict]:
        """Retrieve details about API tokens.

        ??? info "Endpoint"
            [GET /v1/tokens](https://api.prod.generalsystem.com/docs/api#/Identity%20Management%20(v1)/get_v1_tokens)

        Parameters
        ----------
        include_expired:
            set to `True` to include expired tokens in the list.  Defaults to False.

        Returns
        -------
        tokens:
            a list of API tokens.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)
        tokens = dfi.identities.get_tokens(include_expired=False)
        ```
        """
        params = {"includeExpired": str(include_expired).lower()}
        with self.conn.api_get("v1/tokens", params=params, stream=False) as response:
            response.raise_for_status()
            tokens: list[dict] = response.json()
            return tokens

    def create_token(self, name: str, validity: str) -> dict:
        r"""Generate a new API token for the current user.

        ??? info "Endpoint"
            [POST /v1/tokens](https://api.prod.generalsystem.com/docs/api#/Identity%20Management%20(v1)/post_v1_tokens)

        Parameters
        ----------
        name:
            name for token.
        validity:
            how long the token is valid for.  Period of token validity in ISO8601 format. Default is 1 year.

            - pattern: `^P(?:(\d+)Y)?(?:(\d+)M)?(?:(\d+)D)?(?:T(?:(\d+)H)?(?:(\d+)M)?(?:(\d+(?:\.\d{1,3})?)S)?)?$`
            - example: `"P1Y"`

        Returns
        -------
        token info:
            information about the new token.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)

        name = "token_1"
        validity = "P1Y"
        token_info = dfi.identities.create_token(name, validity)
        token = token_info["token]
        ```
        """
        payload = {"name": name, "validity": validity}
        with self.conn.api_post("v1/tokens", json=payload, stream=False) as response:
            response.raise_for_status()
            token_info: dict = response.json()
            return token_info

    def expire_token(self, token_id: str) -> None:
        """Expires the given api token.

        ??? info "Endpoint"
            [DELETE /v1/tokens/{id}](https://api.prod.generalsystem.com/docs/api#/Identity%20Management%20(v1)/delete_v1_identities__id_)

        ??? question "Finding an expired token"
            It's previous existence can still be found with `dfi.get_tokens(include_expired=True)`

        Parameters
        ----------
        token_id:
            the Token ID to expire.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)

        dfi.identities.expire_token("<token id>")
        ```
        """
        with self.conn.api_delete(f"v1/tokens/{token_id}", stream=False) as response:
            response.raise_for_status()
            return None

    def get_identities(self) -> list[dict[str, Any]]:
        """Retrieve a list of identities.

        ??? info "Endpoint"
            [GET /v1/identities](https://api.prod.generalsystem.com/docs/api#/Identity%20Management%20(v1)/get_v1_identities)

        ??? tip "Admin Request"
            You need to be an admin for this request.

        Returns
        -------
        identities:
            a list of identities

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)

        identities = dfi.identities.get_identities()
        ```
        """
        with self.conn.api_get("v1/identities", stream=False) as response:
            response.raise_for_status()
            identities: list[dict[str, Any]] = response.json()
            return identities

    def get_my_identity(self) -> dict[str, Any]:
        """Retrieve data about the identity who made this request.

        ??? info "Endpoint"
            [GET /v1/identities/me](https://api.prod.generalsystem.com/docs/api#/Identity%20Management%20(v1)/get_v1_identities_me)

        Returns
        -------
        identity:
            the caller's identity

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)

        identity = dfi.identities.get_my_identity()
        ```
        """
        with self.conn.api_get("v1/identities/me", stream=False) as response:
            response.raise_for_status()
            identity: dict[str, Any] = response.json()
            return identity

    def get_identity(self, identity_id: str) -> dict[str, Any]:
        """Retrieve data about a specific identity.

        ??? info "Endpoint"
            [GET /v1/identities/{id}](https://api.prod.generalsystem.com/docs/api#/Identity%20Management%20(v1)/get_v1_identities__id_)

        ??? tip "Admin Request"
            You need to be an admin for this request.

        Parameters
        ----------
        identity_id:
            an id of an identity.

        Returns
        -------
        identity:
            the identity of the given id.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)

        identity = dfi.identities.get_identity("<identity id>")
        ```
        """
        with self.conn.api_get(
            f"v1/identities/{identity_id}", stream=False
        ) as response:
            response.raise_for_status()
            identity: dict[str, Any] = response.json()
            return identity

    def _delete_identity(self, identity_id: str) -> None:
        """Remove an identity and all opaque tokens.

        ??? info "Endpoint"
            [DELETE /v1/identities/{id}](https://api.prod.generalsystem.com/docs/api#/Identity%20Management%20(v1)/delete_v1_identities__id_)

        ??? attention
            This should be done in conjunction with removing the user.

        ??? warning
            Deleting the sole identity for a user makes the user unsearchable,
            which makes finding the id to then delete the user impossible.

        ??? tip "Admin Request"
            You need to be an admin for this request.

        Parameters
        ----------
        identity_id:
            the identity id to be deleted.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)

        identity = dfi.identities.delete_identity("<identity id>")
        ```
        """
        with self.conn.api_delete(
            f"v1/identities/{identity_id}", stream=False
        ) as response:
            response.raise_for_status()
            return None

    def get_user_id(self, identity_id: str) -> str:
        """Extract the User ID from an Identity ID.

        Parameters
        ----------
        identity_id:
            the Identity ID to extract the matching User ID from.

        Returns
        -------
        user_id:
            the User ID associated to the given Identity ID

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)

        user_id = dfi.identities.get_user_id("<identity id>")
        ```
        """
        return identity_id.split("|")[1]



"""A QueryDocument describes a query."""

from typing import Any

from typing_extensions import Self

from dfi.errors import InvalidQueryDocument
from dfi.models.filters import FilterField, Only, TimeRange
from dfi.models.filters.geometry import BBox, Polygon
from dfi.models.returns import Count, Records

FILTERS = "filters"
TIME = "time"
MIN_TIME = "minTime"
MAX_TIME = "maxTime"
RETURN_MODEL = "return"
COORDINATES = "coordinates"
FILTER_FIELDS = "fields"
DATASET_ID = "datasetId"
IDS = "id"
GEOMETRY = "geo"
ONLY = "only"


class QueryDocument:
    """A high level interface for declaring a DFI API Query."""

    def __init__(
        self,
        dataset_id: str,
        return_model: Records | Count,
        uids: list[str | int] | None = None,
        geometry: Polygon | BBox | None = None,
        time_range: TimeRange | None = None,
        filter_fields: list[FilterField] | None = None,
        only: Only | str | None = None,
    ) -> None:
        """Initialize a QueryDocument."""
        self._dataset_id: str
        self._return_model: Records | Count
        self._uids: list[str | int] | None = None
        self._geometry: Polygon | BBox | None = None
        self._time_range: TimeRange | None = None
        self._only: Only | None = None
        self._filter_fields: list[FilterField] | None = None

        self._document: dict[str, Any] = {"filters": {}}
        self.set_dataset_id(dataset_id)
        self.set_return_model(return_model)
        self.set_uids(uids)
        self.set_geometry(geometry)
        self.set_time_range(time_range)
        self.set_only(only)
        self.set_filter_fields(filter_fields)

        self.validate()

    def __repr__(self) -> str:
        """Class representation."""
        return f"""QueryDocument(
                dataset_id={self._dataset_id},
                uids={self._uids},
                geometry={self._geometry},
                time_range={self._time_range},
                filter_fields={self._filter_fields},
                only={self._only},
                return_model={self._return_model},
            )"""

    def __str__(self) -> str:
        """Class string formatting."""
        return f"""QueryDocument(
                dataset_id={self._dataset_id},
                uids={self._uids},
                geometry={self._geometry},
                time_range={self._time_range},
                filter_fields={self._filter_fields},
                only={self._only},
                return_model={self._return_model},
            )"""

    def validate(self) -> Self:
        """Validate the QueryDocument.

        Raises
        ------
        InvalidQueryDocument
        """
        self._validate_dataset_id(self._dataset_id)
        self._validate_return_model(self._return_model)
        self._validate_only_filter(self._only, self._return_model)

        return self

    @staticmethod
    def _validate_dataset_id(dataset_id: str) -> None:
        """Check that a dataset_id is set and is a string.

        Parameters
        ----------
        dataset_id:
            a dataset id.
        """
        match dataset_id:
            case str():
                pass
            case _:
                raise InvalidQueryDocument("QueryDocument must have a dataset_id.")

    @staticmethod
    def _validate_return_model(return_model: Records | Count) -> None:
        """Check that a return_model is set and is a Records | Count.

        Parameters
        ----------
        return model:
            How results should be returned.
        """
        match return_model:
            case Records() | Count():
                pass
            case _:
                raise InvalidQueryDocument("QueryDocument must have a return_model.")

    @staticmethod
    def _validate_only_filter(only: Only | str | None, return_model: Records | Count) -> None:
        """Check that an only filter and return_model combination is valid.

        Parameters
        ----------
        only:
            Only filter.
        return model:
            How results should be returned.

        Returns
        -------
        None
        """
        match only, return_model:
            case None, Records() | Count():
                pass
            case Only() | str(), Records():
                pass
            case Only() | str(), Count():
                raise InvalidQueryDocument(f"'{only}' filter is only valid combined with a 'records' return_model.")
            case _, Records() | Count():
                raise InvalidQueryDocument(f"'{only}' is not a valid 'Only' filter.")
            case _:
                raise InvalidQueryDocument(f"'{only}' filter is only valid combined with a 'records' return_model.")

    def build(self) -> dict:
        """Return a formatted Query Document.

        Returns
        -------
        document:
            formatted  Query Document
        """
        self.validate()
        return dict(sorted(self._document.items(), key=lambda x: x[0]))

    @property
    def dataset_id(self) -> str | None:
        """The dataset_id property."""
        return self._dataset_id

    def set_dataset_id(self, dataset_id: str) -> Self:
        """Set the dataset_id property."""
        self._validate_dataset_id(dataset_id)

        match dataset_id:
            case str():
                self._dataset_id = dataset_id
                self._document[DATASET_ID] = dataset_id
            case _:
                raise ValueError("dataset_id must be of type str.")

        return self

    @property
    def return_model(self) -> Records | Count:
        """The return_model property."""
        return self._return_model

    def set_return_model(self, return_model: Records | Count) -> Self:
        """Set the return_model property."""
        self._validate_return_model(return_model)
        self._validate_only_filter(self._only, return_model)

        match return_model:
            case Records() | Count():
                self._return_model = return_model
                self._document[RETURN_MODEL] = return_model.build()
            case _:
                raise ValueError("return_model must be of type ReturnModel.")

        return self

    @property
    def uids(self) -> list[str | int] | None:
        """The uids property."""
        return self._uids

    def set_uids(self, uids: list[str | int] | None) -> Self:
        """Set the uids property."""
        match uids:
            case list():
                self._uids = uids
                self._document[FILTERS][IDS] = uids
            case None:
                self._uids = uids
                self._document[FILTERS].pop(IDS, None)
            case _:
                raise ValueError("uids must be of type list[str | int].")

        return self

    @property
    def only(self) -> Only | None:
        """The only property."""
        return self._only

    def set_only(self, only: Only | str | None) -> Self:
        """Set the only property."""
        self._validate_only_filter(only, self._return_model)

        match only:
            case Only():
                self._only = only
                self._document[FILTERS][ONLY] = self._only.build()
            case str():
                self._only = Only(only)
                self._document[FILTERS][ONLY] = self._only.build()
            case None:
                self._only = only
                self._document[FILTERS].pop(ONLY, None)
            case _:
                raise ValueError("only must be of type Only | None.")

        return self

    @property
    def time_range(self) -> TimeRange | None:
        """The time_range property."""
        return self._time_range

    def set_time_range(self, time_range: TimeRange | None) -> Self:
        """Set the time_range property."""
        match time_range:
            case TimeRange():
                self._time_range = time_range
                self._document[FILTERS][TIME] = time_range.build()
            case None:
                self._time_range = time_range
                self._document[FILTERS].pop(TIME, None)
            case _:
                raise ValueError("time_range must be of type TimeRange | None.")

        return self

    @property
    def geometry(self) -> Polygon | BBox | None:
        """The geometry property."""
        return self._geometry

    def set_geometry(self, geometry: Polygon | BBox | None) -> Self:
        """Set the geometry property."""
        match geometry:
            case Polygon() | BBox():
                self._geometry = geometry
                self._document[FILTERS][GEOMETRY] = geometry.build()
            case None:
                self._geometry = geometry
                self._document[FILTERS].pop(GEOMETRY, None)
            case _:
                raise ValueError("geometry must be of type Polygon | BBox | None.")

        return self

    @property
    def filter_fields(self) -> list[FilterField] | None:
        """The filter_fields property."""
        return self._filter_fields

    def set_filter_field(self, filter_field: FilterField) -> Self:
        """Set a filter field.

        - An existing FilterField will be overwritten by a new FilterField with the same name.
        - Possibility that more fields can be set than are allowed.  The API will catch this error.
        """
        match filter_field:
            case FilterField():
                if self._document[FILTERS].get(FILTER_FIELDS) is None:
                    self._document[FILTERS][FILTER_FIELDS] = {}

                self._document[FILTERS][FILTER_FIELDS].update(filter_field.build())
            case _:
                raise ValueError("filter_field must be of type FilterField.")

        return self

    def set_filter_fields(self, filter_fields: list[FilterField] | None) -> Self:
        """Set a filter field.

        - An existing FilterField will be overwritten by a new FilterField with the same name.
        - Possibility that more fields can be set than are allowed.  The API will catch this error.

        Parameters
        ----------
        filter_fields:

            - If `list[FilterField]`, will iteratively add each field.
            - If `None`, will delete all existing filter fields.
        """
        match filter_fields:
            case list():
                for field in filter_fields:
                    self.set_filter_field(field)
            case None:
                self._filter_fields = None
                self._document[FILTERS].pop(FILTER_FIELDS, None)
            case _:
                raise ValueError("filter_field must be of type list[FilterField] | None.")

        return self



"""Class composition of all the other classes, to access the wrappers with syntactic sugar."""

from dfi.connect import Connect
from dfi.services.datasets import Datasets
from dfi.services.identities import Identities
from dfi.services.info import Info
from dfi.services.ingest import Ingest
from dfi.services.query import Query
from dfi.services.users import Users


class Client:
    """Collection of GS Platform services."""

    def __init__(
        self,
        api_token: str,
        base_url: str | None = None,
        query_timeout: int | None = 60,
        progress_bar: bool | None = False,
    ) -> None:
        """Create a connection to GS Platform services.

        Parameters
        ----------
        api_token:
            a unique token.
        base_url:
            where the GS Platform is located.
        query_timeout:
            will timeout if no response within this many seconds.
        progress_bar:
            if `True`, will show a progress bar.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)
        ```
        """
        self.conn = Connect(
            api_token=api_token,
            base_url=base_url,
            query_timeout=query_timeout,
            progress_bar=progress_bar,
        )
        self.datasets = Datasets(self.conn)
        self.identities = Identities(self.conn)
        self.info = Info(self.conn)
        self.ingest = Ingest(self.conn)
        self.users = Users(self.conn)
        self.query = Query(self.conn)

    def __repr__(self) -> str:
        """Class representation."""
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        """Class string formatting."""
        return f"""Instance of dfi.{self.__class__.__name__} composed with: {self.conn!s}"""



"""Allows access to submodules from dfi namespace."""

from __future__ import annotations

from importlib.metadata import version

__version__ = version("dfipy")

from dfi.client import Client  # noqa: F401 (unused-import)



"""Class for querying for data with DFI Query V1 API."""

import json  # noqa: I001
import logging
from typing import Any

import pandas as pd
from sseclient import SSEClient
from tqdm import tqdm

from dfi.connect import Connect
from dfi.errors import (
    DFIResponseError,
    EventsMissedError,
    NoEventsRecievedError,
    NoFinishMessageReceivedError,
    UnkownMessageReceivedError,
    TimeRangeUndefinedError,  # noqa: F401 (import allows hyperlink to definition in docstring)
    PolygonUndefinedError,  # noqa: F401 (import allows hyperlink to definition in docstring)
    BBoxUndefinedError,  # noqa: F401 (import allows hyperlink to definition in docstring)
)
from dfi.models import QueryDocument
from dfi.models.filters import FilterField, Only, TimeRange
from dfi.models.filters.geometry import BBox, Polygon
from dfi.models.returns import Count, GroupBy, IncludeField, Records

_logger = logging.getLogger(__name__)


class Query:
    """Class responsible for requests to the Query V1 DFI API.

    It can be accessed via the a dfi.Client class instance or it must be instantiated
    with a dfi.Connect instance as argument.
    """

    def __init__(self, conn: Connect) -> None:
        """Handle queries to DFI Query V1 API.

        Parameters
        ----------
        conn:
            a Connect instance.
        """
        self.conn = conn
        self._document: dict | None = None

    def __repr__(self) -> str:
        """Class representation."""
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        """Class string formatting."""
        return f"{self.__class__.__name__}(conn={self.conn!r})"

    @property
    def document(self) -> dict | None:
        """The Query Document used in the last query for data."""
        return self._document

    def instrumentation(
        self,
        dataset_id: str | None = None,
        identity_id: str | None = None,
        before: str | None = None,
        page_size: int | None = None,
    ) -> list[dict]:  # type: ignore[type-arg]
        """Retrieve a list of queries made to datasets.

        By default this will return a list of the 100 most recent queries to any dataset.
        Information about the queries can be retrieved by utilizing the `before` and `page_size` parameters
        to select a page of information.

        ??? info "Endpoint"
            [GET /v1/query/instrumentation](https://api.prod.generalsystem.com/docs/api#/Query%20(v1)/get_v1_query_instrumentation)

        ??? tip "Tenant Admins"
            Tenant admins will be able to see any query ran on any dataset in their tenant.
            - filter by dataset
            - filter by identity

        ??? tip "Non-Admins"
            Non-admins will only see their own queries.
            - filter by dataset

        Parameters
        ----------
        dataset_id:
            Filter results to only include this dataset.
        identity_id:
            Filter results to only include this identity.
        before:
            [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) string. Only retrieve items created before this given time. Defaults to now.
        page_size:
            Number of items to return in the response. Maximum is 500. Default is 100 if omitted.

        Raises
        ------
        DFIResponseError

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client("<token>", "<url>")

        dfi.query.instrumentation()
        ```
        ```python
        [
            {
                "identityId": "user|0ffb434d-f319-463a-8cf1-6ff939244486",
                "datasetId": "gs.dfi",
                "apiVersion": "5.0.2",
                "responseType": "count",
                "result": "success",
                "idCount": 0,
                "timeRange": None,
                "fieldCount": 0,
                "polygonSize": 0,
                "polygonArea": 0,
                "ingressTime": 1708544552000,
                "issuedElapsed": 8,
                "firstByteElapsed": 154,
                "lastByteElapsed": 154,
            },
            {
                "identityId": "user|0ffb434d-f319-463a-8cf1-6ff939244486",
                "datasetId": "gs.test-dataset",
                "apiVersion": "5.0.2",
                "responseType": "items",
                "result": "success",
                "idCount": 0,
                "timeRange": None,
                "fieldCount": 0,
                "polygonSize": 0,
                "polygonArea": 0,
                "ingressTime": 1708541136914,
                "issuedElapsed": 40,
                "firstByteElapsed": 154,
                "lastByteElapsed": 412,
            },
            {
                "identityId": "user|0ffb434d-f319-463a-8cf1-6ff939244486",
                "datasetId": "gs.test-dataset",
                "apiVersion": "5.0.2",
                "responseType": "itemsWithoutPayload",
                "result": "success",
                "idCount": 0,
                "timeRange": None,
                "fieldCount": 0,
                "polygonSize": 0,
                "polygonArea": 0,
                "ingressTime": 1708541108892,
                "issuedElapsed": 11,
                "firstByteElapsed": 156,
                "lastByteElapsed": 255,
            },
            {
                "identityId": "user|0ffb434d-f319-463a-8cf1-6ff939244486",
                "datasetId": "gs.test-dataset",
                "apiVersion": "5.0.2",
                "responseType": "entities",
                "result": "success",
                "idCount": 1,
                "timeRange": 2678400000,
                "fieldCount": 1,
                "polygonSize": 5,
                "polygonArea": 12391399902,
                "ingressTime": 1708540985751,
                "issuedElapsed": 8,
                "firstByteElapsed": 151,
                "lastByteElapsed": 151,
            },
        ]
        ```
        """
        params = {
            "identityId": identity_id,
            "datasetId": dataset_id,
            "before": before,
            "pageSize": page_size,
        }
        with self.conn.api_get(
            "v1/query/instrumentation", params=params, stream=False
        ) as response:
            response.raise_for_status()
            return response.json()

    def manage(self, dataset_id: str, operation: str) -> dict[str, str]:  # type: ignore[type-arg]
        """Run a data management query.

        This allows you to perform data management operations on a dataset.
        **All operations are irrevokable and cannot be undone.**

        ??? info "Endpoint"
            [POST /v1/query/manage](https://api.prod.generalsystem.com/docs/api#/Query%20(v1)/post_v1_query_manage)

        ??? tip "Admin Request"
            You need to be an admin for this request.

        Parameters
        ----------
        dataset_id:
            Filter results to only include this dataset.
        operation:
            The operation to perform.  Supported operations are

            - `truncate` - Remove all data from this dataset

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client("<token>", "<url>")

        dataset_id = "<dataset id>"
        operation = "truncate"
        dfi.query.manage(dataset_id, operation)
        ```
        ```python
        {
            "status": "success"
        }
        ```
        """
        body = {
            "datasetId": dataset_id,
            "operation": operation,
        }
        with self.conn.api_post("v1/query/manage", json=body, stream=False) as response:
            response.raise_for_status()
            return response.json()

    def count(
        self,
        dataset_id: str,
        uids: list[str | int] | None = None,
        geometry: Polygon | BBox | None = None,
        time_range: TimeRange | None = None,
        filter_fields: list[FilterField] | None = None,
    ) -> int:
        """Query for the number of records within the filter bounds.

        ??? info "Endpoint"
            [POST /v1/query](https://api.prod.generalsystem.com/docs/api#/Query%20(v1)/post_v1_query)

        Parameters
        ----------
        dataset_id:
            the dataset to be queried.
        uids:
            specifies which uids to search for.
        geometry:
            specifies the spatial bounds to search within.
        time_range:
            specifies the time bounds to search within.
        filter_fields:
            specifies filters on Filter Fields.

        Returns
        -------
        count:
            The number of records stored in the DFI engine.

        Raises
        ------
        DFIResponseError
        TimeRangeUndefinedError
        PolygonUndefinedError
        BBoxUndefinedError
        ValueError

        Examples
        --------
        ```python
        from dfi import Client
        from dfi.models.filters import TimeRange
        from dfi.models.filters.geometry import Polygon

        dfi = Client("<token>", "<url>")

        dataset_id = "<dataset id>"

        uids = ["01234567-89AB-CDEF-1234-0123456789AB"]
        time_range = TimeRange().from_strings(
            min_time="2022-01-01T08:00:00Z",
            max_time="2022-02-01T08:00:00Z",
        )

        coordinates = [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 1.0],
            [0.0, 0.0],
        ]
        geometry = Polygon().from_raw_coords(
            coordinates, geojson=True
        )

        dfi.query.count(
            dataset_id,
            uids=uids,
            time_range=time_range,
            geometry=geometry,
        )
        ```
        """
        query_doc = QueryDocument(
            dataset_id=dataset_id,
            return_model=Count(),
            uids=uids,
            time_range=time_range,
            geometry=geometry,
            filter_fields=filter_fields,
        )
        self._document = query_doc.build()

        with self.conn.api_post("v1/query", json=self._document) as response:
            client = SSEClient(response)  # type: ignore[arg-type]
            return self._receive_counts(client)

    def unique_id_counts(
        self,
        dataset_id: str,
        uids: list[str | int] | None = None,
        geometry: Polygon | BBox | None = None,
        time_range: TimeRange | None = None,
        filter_fields: list[FilterField] | None = None,
    ) -> dict[str | int, int]:
        """Query for the number of records for each id within the filter bounds.

        ??? info "Endpoint"
            [POST /v1/query](https://api.prod.generalsystem.com/docs/api#/Query%20(v1)/post_v1_query)

        Parameters
        ----------
        dataset_id:
            the dataset to be queried.
        uids:
            specifies which uids to search for.
        geometry:
            specifies the spatial bounds to search within.
        time_range:
            specifies the time bounds to search within.
        filter_fields:
            specifies filters on Filter Fields.

        Returns
        -------
        unique ids count
            The count of records for each id within the bounds.

        Raises
        ------
        DFIResponseError
        TimeRangeUndefinedError
        PolygonUndefinedError
        BBoxUndefinedError
        ValueError

        Examples
        --------
        ```python
        from dfi import Client
        from dfi.models.filters import TimeRange
        from dfi.models.filters.geometry import Polygon

        dfi = Client("<token>", "<url>")

        dataset_id = "<dataset id>"

        uids = ["01234567-89AB-CDEF-1234-0123456789AB"]
        time_range = TimeRange().from_strings(
            min_time="2022-01-01T08:00:00Z",
            max_time="2022-02-01T08:00:00Z",
        )

        coordinates = [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 1.0],
            [0.0, 0.0],
        ]
        polygon = Polygon().from_raw_coords(
            coordinates, geojson=True
        )

        dfi.query.unique_id_counts(
            dataset_id,
            uids=uids,
            time_range=time_range,
            geometry=geometry,
        )
        ```
        """
        query_doc = QueryDocument(
            dataset_id=dataset_id,
            return_model=Count(groupby=GroupBy("uniqueId")),
            uids=uids,
            time_range=time_range,
            geometry=geometry,
            filter_fields=filter_fields,
        )
        self._document = query_doc.build()

        with self.conn.api_post("v1/query", json=self._document) as response:
            client = SSEClient(response)  # type: ignore[arg-type]
            return self._receive_unique_id_counts(client)

    def records(
        self,
        dataset_id: str,
        uids: list[str | int] | None = None,
        geometry: Polygon | BBox | None = None,
        time_range: TimeRange | None = None,
        only: Only | str | None = None,
        filter_fields: list[FilterField] | None = None,
        include: list[IncludeField | str] | None = None,
    ) -> pd.DataFrame:
        """Query for the records within the filter bounds.

        ??? info "Endpoint"
            [POST /v1/query](https://api.prod.generalsystem.com/docs/api#/Query%20(v1)/post_v1_query)

        Parameters
        ----------
        dataset_id:
            the dataset to be queried.
        uids:
            specifies which uids to search for.
        geometry:
            specifies the spatial bounds to search within.
        time_range:
            specifies the time bounds to search within.
        only:
            specifies that only the newest or oldest record is retuned.
        filter_fields:
            specifies filters on Filter Fields.
        include:
            specifies the extra fields to include in the returned results.

        Returns
        -------
        records
            The count of records for each id within the bounds.

        Raises
        ------
        DFIResponseError
        TimeRangeUndefinedError
        PolygonUndefinedError
        BBoxUndefinedError
        ValueError

        Examples
        --------
        ```python
        from dfi import Client
        from dfi.models.filters import Only, TimeRange
        from dfi.models.filters.geometry import Polygon

        dfi = Client("<token>", "<url>")

        dataset_id = "<dataset id>"

        uids = ["01234567-89AB-CDEF-1234-0123456789AB"]
        time_range = TimeRange().from_strings(
            min_time="2022-01-01T08:00:00Z",
            max_time="2022-02-01T08:00:00Z",
        )

        coordinates = [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 1.0],
            [0.0, 0.0],
        ]
        polygon = Polygon().from_raw_coords(
            coordinates, geojson=True
        )

        dfi.query.records(
            dataset_id,
            uids=uids,
            time_range=time_range,
            geometry=geometry,
            only="newest",
            include=["fields", "metadataId"],
        )
        ```
        """
        query_doc = QueryDocument(
            dataset_id=dataset_id,
            return_model=Records(include=include),
            uids=uids,
            time_range=time_range,
            geometry=geometry,
            only=only,
            filter_fields=filter_fields,
        )
        self._document = query_doc.build()

        with self.conn.api_post("v1/query", json=self._document) as response:
            client = SSEClient(response)  # type: ignore[arg-type]
            return self._receive_records(client)

    def raw_request(self, document: dict[str, Any]) -> pd.DataFrame | list[str] | int:
        """Provide an escape hatch for those who definitely, absolutely, 100% know what they're doing.

        No validation of the query document is done before saying sending off the request.
        The "hold my beer" of queries.

        ??? info "Endpoint"
            [POST /v1/query](https://api.prod.generalsystem.com/docs/api#/Query%20(v1)/post_v1_query)

        Parameters
        ----------
        document:
            The full request body for POST /v1/query.
        """
        self._document = document

        with self.conn.api_post("v1/query", json=document) as response:
            client = SSEClient(response)  # type: ignore[arg-type]
            match document:
                case {"return": {"type": "count", "groupBy": {"type": "uniqueId"}}}:
                    return self._receive_unique_id_counts(client)
                case {"return": "count"} | {"return": {"type": "count"}}:
                    return self._receive_counts(client)
                case {"return": {"type": "records"}}:
                    return self._receive_records(client)
                case _:
                    # This will only happen in DFI API adds new return types before dfipy is updated.
                    raise ValueError("Unknown return type.")

    def _receive_counts(self, client: SSEClient) -> int:
        """Collect 'count' results by summing the results received.

        Parameters
        ----------
        client:
            SSE client for response.

        Raises
        ------
        DFIResponseError
        UnkownMessageReceivedError
        NoEventsRecievedError
        NoFinishMessageReceivedError
        EventsMissedError
        """
        counts = 0

        events_list_is_empty = True
        finish_message = False
        messages_received = 0

        for event in (
            pbar := tqdm(
                client.events(),
                disable=not self.conn.progress_bar,
                maxinterval=0.5,
                miniters=1,
            )
        ):
            events_list_is_empty = False

            match event.event:
                case "keepAlive":
                    continue
                case "message":
                    messages_received += 1
                    counts += json.loads(event.data)
                    pbar.set_description(f"Collecting {counts:,} counts")
                    continue
                case "finish":
                    finish_message = True
                    messages_sent = json.loads(event.data).get("messageCount")
                    break
                case "queryError":
                    raise DFIResponseError(event.data)
                case _:
                    raise UnkownMessageReceivedError(event)

        if events_list_is_empty:
            raise NoEventsRecievedError("0 events received from DFI API.")

        if not finish_message:
            raise NoFinishMessageReceivedError(
                "Stream ended before finish message received.  Results may not be complete."
            )

        if messages_received != messages_sent:
            raise EventsMissedError(
                f"Received {messages_received}/{messages_sent} events from DFI API."
            )

        return counts

    def _receive_unique_id_counts(self, client: SSEClient) -> dict[str | int, int]:
        """Collect 'uniqueIds groupBy' results.

        Parameters
        ----------
        client:
            SSE client for response.

        Raises
        ------
        DFIResponseError
        UnkownMessageReceivedError
        NoEventsRecievedError
        NoFinishMessageReceivedError
        EventsMissedError
        """
        unique_id_counts = {}
        events_list_is_empty = True
        finish_message = False
        messages_received = 0

        for event in (
            pbar := tqdm(
                client.events(),
                disable=not self.conn.progress_bar,
                maxinterval=0.5,
                miniters=1,
            )
        ):
            events_list_is_empty = False

            match event.event:
                case "keepAlive":
                    continue
                case "message":
                    messages_received += 1
                    unique_id_counts.update(json.loads(event.data))
                    pbar.set_description(
                        f"Collecting {len(unique_id_counts):,} id counts."
                    )
                    continue
                case "finish":
                    finish_message = True
                    messages_sent = json.loads(event.data).get("messageCount")
                    break
                case "queryError":
                    raise DFIResponseError(event.data)
                case _:
                    raise UnkownMessageReceivedError(event)

        if events_list_is_empty:
            raise NoEventsRecievedError("0 events received from DFI API.")

        if not finish_message:
            raise NoFinishMessageReceivedError(
                "Stream ended before finish message received.  Results may not be complete."
            )

        if messages_received != messages_sent:
            raise EventsMissedError(
                f"Received {messages_received}/{messages_sent} events from DFI API."
            )

        return unique_id_counts

    def _receive_records(self, client: SSEClient) -> pd.DataFrame:
        """Collect 'records' results into Pandas DataFrame.

        Parameters
        ----------
        client:
            SSE client for response.

        Raises
        ------
        DFIResponseError
        UnkownMessageReceivedError
        NoEventsRecievedError
        NoFinishMessageReceivedError
        EventsMissedError
        """
        records = []
        events_list_is_empty = True
        finish_message = False
        messages_received = 0

        for event in (
            pbar := tqdm(
                client.events(),
                disable=not self.conn.progress_bar,
                maxinterval=0.5,
                miniters=1,
            )
        ):
            events_list_is_empty = False

            match event.event:
                case "keepAlive":
                    continue
                case "message":
                    messages_received += 1
                    records += json.loads(event.data)
                    pbar.set_description(f"Collecting {len(records):,} records.")
                    continue
                case "finish":
                    finish_message = True
                    messages_sent = json.loads(event.data).get("messageCount")
                    break
                case "queryError":
                    raise DFIResponseError(event.data)
                case _:
                    raise UnkownMessageReceivedError(event)

        if events_list_is_empty:
            raise NoEventsRecievedError("0 events received from DFI API.")

        if not finish_message:
            raise NoFinishMessageReceivedError(
                "Stream ended before finish message received.  Results may not be complete."
            )

        if messages_received != messages_sent:
            raise EventsMissedError(
                f"Received {messages_received}/{messages_sent} events from DFI API."
            )

        if len(records) > 0:
            return pd.DataFrame(records).assign(time=lambda df: pd.to_datetime(df.time))
        else:
            return pd.DataFrame(columns=["id", "coordinate", "time"])



"""Validation and error formatting for DFI responses."""

import json
import logging

import requests

from dfi.errors import DFIResponseError

_logger = logging.getLogger(__name__)

SUCCESS_CODES = 2


def response(
    resp: requests.models.Response,
    url: str,
    headers: dict,
    params: dict | None = None,
    payload: dict | list | None = None,
) -> None:
    """Log the response of a request with the given parameters. Raise an error if status code is not 20x.

    Parameters
    ----------
    resp:
        a response object
    url:
        the queried url
    headers:
        request headers
    params:
        request params
    payload:
        request payload

    Raises
    ------
    DFIResponseError
        If there was an error querying the DFI API.
    """
    # prevent from showing the user token to terminal and logs
    headers = headers.copy()
    headers["Authorization"] = "Bearer XXX"

    msg = f"""
STATUS CODE: {resp.status_code}
ERROR: {resp.text}
URL: {url}
HEADERS: {json.dumps(headers, sort_keys=True, indent=4)}
PARAMS: {json.dumps(params, sort_keys=True, indent=4)}
"""
    if payload is not None:
        msg += f"PAYLOAD: {json.dumps(payload, sort_keys=True, indent=4)}"
    else:
        msg += f"PAYLOAD: {json.dumps(None)}"

    if int(resp.status_code // 100) != SUCCESS_CODES:  # any code 2xx is a valid success response code
        _logger.error(msg)
        raise DFIResponseError(msg)

    _logger.debug(msg)



"""Models module."""

from typing import Any, TypeAlias

from dfi.models.query_document import QueryDocument  # noqa: F401 (unused-import)

FilterFields: TypeAlias = dict[str, dict[str, Any]]
"""Alias for `dict[str, dict[str, Any]]`"""



"""TimeRange filter describes how to filter on time."""

from datetime import datetime, timezone

from typing_extensions import Self

from dfi.errors import TimeRangeMismatchError, TimeRangeUndefinedError, TimeZoneUndefinedError

# DATETIME_MIN = 0
# DATETIME_MAX = 2_147_483_647


class TimeRange:
    """A TimeRange is a a way to bound time.

    If a bound (`min_time` or `max_time`) is None, then time in that direction is unbound
    i.e. bounds will "expand" to the first or last record in dataset based on timestamp.
    """

    _min_time: datetime | None
    _max_time: datetime | None

    def __repr__(self) -> str:
        """Class representation."""
        time_range = self.build()
        min_time = time_range["minTime"]
        max_time = time_range["maxTime"]
        return f"TimeRange({min_time}, {max_time})"

    def __str__(self) -> str:
        """Class string formatting."""
        time_range = self.build()
        min_time = time_range["minTime"]
        max_time = time_range["maxTime"]
        return f"TimeRange({min_time}, {max_time})"

    @property
    def min_time(self) -> datetime | None:
        """The min_time property."""
        return self._min_time

    @property
    def max_time(self) -> datetime | None:
        """The max_time property."""
        return self._max_time

    def from_datetimes(self, min_time: datetime | None = None, max_time: datetime | None = None) -> Self:
        """Create a TimeRange from `datetime` objects.

        Parameters
        ----------
        min_time:
            Lower time bound.
        max_time:
            Upper time bound.

        Returns
        -------
        TimeRange

        Raises
        ------
        TimeRangeUndefinedError
        TimeRangeMismatchError
        TypeError

        Examples
        --------
        ### TimeRange From `datetime`s
        ```python
        from datetime import datetime, timezone

        min_time = datetime(
            2020, 1, 1, 0, 0, 0, tzinfo=timezone.utc
        )
        min_time = datetime(
            2020, 1, 1, 0, 0, 1, tzinfo=timezone.utc
        )

        TimeRange().from_datetimes(min_time, max_time)
        ```
        ```python
        TimeRange(2020-01-01T00:00:00, 2020-01-01T00:00:01)
        ```
        """
        match min_time:
            case datetime():
                self._min_time = min_time
            case None:
                self._min_time = None
            case _:
                raise TypeError(f"min_time should be of type str | None, found '{type(min_time)}'")

        match max_time:
            case datetime():
                self._max_time = max_time
            case None:
                self._max_time = None
            case _:
                raise TypeError(f"max_time should be of type str | None, found '{type(max_time)}'")

        return self.validate()

    def from_strings(self, min_time: str | None = None, max_time: str | None = None) -> Self:
        """Create a TimeRange from [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) strings.

        Parameters
        ----------
        min_time:
            Lower time bound.
        max_time:
            Upper time bound.

        Returns
        -------
        TimeRange

        Raises
        ------
        TimeRangeUndefinedError
        TimeRangeMismatchError
        TypeError

        Examples
        --------
        ??? attention

            Creating datetimes from strings in Zulu time (e.g. `"2024-02-06T10:41:54Z"`) are not supported in Python 3.10


        ### TimeRange From Strings
        ```python
        TimeRange().from_strings(
            "2020-01-01T00:00:00.001000+00:00",
            "2020-01-01T00:00:01.001000+00:00",
        )
        ```
        ```python
        TimeRange(2020-01-01T00:00:00+01:00, 2020-01-01T00:00:01+01:00)
        ```
        """
        match min_time:
            case str():
                self._min_time = datetime.fromisoformat(min_time)
            case None:
                self._min_time = None
            case _:
                raise TypeError(f"min_time should be of type str | None, found '{type(min_time)}'")

        match max_time:
            case str():
                self._max_time = datetime.fromisoformat(max_time)
            case None:
                self._max_time = None
            case _:
                raise TypeError(f"max_time should be of type str | None, found '{type(max_time)}'")

        return self.validate()

    def from_millis(
        self, min_time: int | None = None, max_time: int | None = None, tz: timezone = timezone.utc
    ) -> Self:
        """Create a TimeRange from integers representing Unix Epoch in milliseconds.

        Parameters
        ----------
        min_time:
            Lower time bound.
        max_time:
            Upper time bound.
        tz:
            The timezone for the datetime.  Defaults to UTC.

        Returns
        -------
        TimeRange

        Raises
        ------
        TimeRangeUndefinedError
        TimeRangeMismatchError
        TypeError

        Examples
        --------
        ### TimeRange From Milliseconds Since Unix Epoch
        ```python
        TimeRange().from_millis(1577836800000, 1577836801000)
        ```
        ```python
        TimeRange(2020-01-01T00:00:00.001000+00:00, 2020-01-01T00:00:01.001000+00:00)
        ```
        """
        match min_time:
            case int():
                self._min_time = datetime.fromtimestamp(min_time / 1000.0, tz)
            case None:
                self._min_time = None
            case _:
                raise TypeError(f"min_time should be of type str | None, found '{type(min_time)}'")

        match max_time:
            case int():
                self._max_time = datetime.fromtimestamp(max_time / 1000.0, tz)
            case None:
                self._max_time = None
            case _:
                raise TypeError(f"max_time should be of type str | None, found '{type(max_time)}'")

        return self.validate()

    def validate(self) -> Self:
        """Validate the TimeRange bounds.

        Raises
        ------
        TimeRangeUndefinedError
        TimeRangeMismatchError
        """
        if not (hasattr(self, "_min_time") or hasattr(self, "_max_time")):
            raise TimeRangeUndefinedError

        if (self._min_time is not None) and (self._max_time is not None):
            if self._min_time > self._max_time:
                raise TimeRangeMismatchError

            if (self._min_time.tzinfo is None) or (self._max_time.tzinfo is None):
                raise TimeZoneUndefinedError

            # TODO ask @matt what the limits (if any) are
            # if self._min_time < 0:
            #     raise TimeRangeOutOfBoundsError(
            #         f"min_time value '{self._min_time}' not within ({DATETIME_MIN}, {DATETIME_MAX})"
            #     )
            # if self._max_time > 0:
            #     raise TimeRangeOutOfBoundsError(
            #         f"max_time value '{self._max_time}' not within ({DATETIME_MIN}, {DATETIME_MAX})"
            #     )

        return self

    def build(self) -> dict[str, str | None]:
        """Validate the TimeRange is defined and format the TimeRange for use in Query Document."""
        self.validate()
        min_time = None if self._min_time is None else self._min_time.isoformat()
        max_time = None if self._max_time is None else self._max_time.isoformat()

        return {"maxTime": max_time, "minTime": min_time}



"""Manages user authentication via the User Management API service."""

from dfi.connect import Connect


class Users:
    """Class responsible handling users."""

    def __init__(self, conn: Connect) -> None:
        self.conn = conn

    def __repr__(self) -> str:
        """Class representation."""
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        """Class string formatting."""
        return f"""Instance of dfi.{self.__class__.__name__} composed with: {self.conn!s}"""

    def create_user(self, user: dict) -> dict:
        """Create a new user.

        ??? info "Endpoint"
            [POST /v1/users](https://api.prod.generalsystem.com/docs/api#/User%20Management%20(v1)/post_v1_users)

        ??? tip "Admin Request"
            You need to be an admin for this request.

        Parameters
        ----------
        user:
            a user specification to create.

        Returns
        -------
        user:
            the created user.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)

        user = {
            "userName": "iain.m.banks",
            "name": {
                "givenName": "Iain",
                "familyName": "Banks",
            },
            "emails": [
                {
                    "value": "iain.m.banks@organization.com",
                    "primary": True,
                }
            ],
            "active": True,
        }
        created_user = dfi.users.create_user(user)
        ```
        """
        with self.conn.api_post("v1/users", json=user, stream=False) as response:
            response.raise_for_status()
            created_user: dict = response.json()
            return created_user

    def get_users(self) -> list[dict]:
        """Retrieve the list of users in the realm.

        ??? info "Endpoint"
            [GET /v1/users](https://api.prod.generalsystem.com/docs/api#/User%20Management%20(v1)/get_v1_users)

        ??? tip "Admin Request"
            You need to be an admin for this request.

        Returns
        -------
        identity:
            the identity of the given id.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)

        users = dfi.users.get_users()
        ```
        """
        with self.conn.api_get("v1/users", stream=False) as response:
            response.raise_for_status()
            users: list[dict] = response.json()
            return users

    def get_user(self, user_id: str) -> dict:
        """Retrieve details about a user by id.

        ??? info "Endpoint"
            [GET /v1/users/{id}](https://api.prod.generalsystem.com/docs/api#/User%20Management%20(v1)/get_v1_users__id_)

        ??? tip "Admin Request"
            You need to be an admin for this request.

        Parameters
        ----------
        user_id:
            an id of a user.

        Returns
        -------
        identity:
            the identity of the given id.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)

        identity = dfi.users.get_user("<user id>")
        ```
        """
        with self.conn.api_get(f"v1/users/{user_id}", stream=False) as response:
            response.raise_for_status()
            user: dict = response.json()
            return user

    def delete_user(self, user_id: str) -> None:
        """Remove a user from the system. This removes all identities and access tokens that exist for the user as well.

        ??? info "Endpoint"
            [DELETE /v1/users/{id}](https://api.prod.generalsystem.com/docs/api#/User%20Management%20(v1)/delete_v1_users__id_)

        ??? tip "Admin Request"
            You need to be an admin for this request.

        Parameters
        ----------
        user_id:
            an id of a user.

        Returns
        -------
        identity:
            the identity of the given id.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)

        identity = dfi.users.delete_user("<user id>")
        ```
        """
        with self.conn.api_delete(f"v1/users/{user_id}", stream=False) as response:
            response.raise_for_status()
            return None



"""Manages ingest of data via the Import API service.

???+ note "Why 'Ingest' instead of 'Import'"
    The name "ingest" is used in place of API name "import" because the word "import" is a reserved name in Python.
"""

import json
from typing import Any

import boto3
import natsort
from botocore.client import Config
from tqdm import tqdm

from dfi.connect import Connect


class AWSCredentials:
    """A constructor class for building an AWSCredentials document.

    Parameters which point to an AWS role that the importer will assume to scan and download files.
    These parameters are a subset of the aws AssumeRoleCommand
    [https://docs.aws.amazon.com/AWSJavaScriptSDK/v3/latest/clients/client-account/classes/assumerolecommand.html](https://docs.aws.amazon.com/AWSJavaScriptSDK/v3/latest/clients/client-account/classes/assumerolecommand.html)
    """

    def __init__(
        self,
        role_arn: str,
        policy: dict | None = None,
        policy_arns: list[str] | None = None,
    ) -> None:
        """Initilaize class and stores values.

        Parameters
        ----------
        role_arn:
            The Amazon Resource Name (ARN) of the AWS role for us to assume.
        policy:
            a JSON policy that we should use when assuming the role.
        policy_arns:
            a list of Policy ARNs to use as managed session policies.

        Examples
        --------
        ```python
        from dfi.services.ingest import AWSCredentials

        role_arn = "arn:aws:s3:::example_bucket/example_key"
        AWSCredentials(role_arn).build()
        ```
        ```python
        {"RoleArn": "arn:aws:s3:::example_bucket/example_key"}
        ```
        """
        self.role_arn = role_arn
        self.policy = policy
        self.policy_arns = policy_arns

    def build(self) -> dict[str, Any]:
        """Build an AWSCredentials document from given inputs."""
        aws_credentials: dict[str, Any] = {"RoleArn": self.role_arn}
        if self.policy:
            aws_credentials.update({"Policy": self.policy})
        if self.policy_arns:
            aws_credentials.update({"PolicyArns": self.policy_arns})

        return aws_credentials


class BatchURLFiles:
    """A constructor class for building an BatchURLFiles document.  Indicates that this import batch should source files from a provided list of URLs."""

    def __init__(self, urls: list[str]) -> None:
        """Initialize and store values.

        Examples
        --------
        In this example, we've generated presigned URLs for 4 files on S3.  These files can now be ingested without
        needing to share AWS credentials.
        ```python
        from dfi.services.ingest import BatchURLFiles

        urls = [
            "https://bucket-name.s3.eu-west-2.amazonaws.com/sample-data/file-1.csv?response-content-disposition=inline&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEOX%2F%2F%2F%2F%2F%2F%2F%2F%2F%2FwEaCWV1LXdlc3QtMiJGMEQCICzap9Nfx0BY9IPhrWCPEWwiR4OYW%2BxxheB%2FeorchP6QAiB5lRoXq3cjr%2BTFMvjuvGoCTaMxw9T4SzbYoUzIDMtIwyqoBAiO%2F%2F%2F%2F%2F%2F%2F%2F%2F%2F8BEAMaDDY2OTM1OTQ2MDU1NCIMAGNCpXF5mRRjsS7mKvwDcOy9VbRwNRJKZFkEdUtcXHRkKTFw1t%2FSYDaqE%2BrdEhV01bV7v5uuEJQye9zI24r%2FHJvS1dolfumoaDdjkkI6%2FtODeGo0WiZsrI8yOxCFL6WCDnEIS%2F9VJpN7MdXR9kdMhPu1crhOM52PmD0QUO6jckS1RoxXyDChrW6xB9dI7w6h%2F%2B674gHg8De%2BAvJFe45eGYqPNogwLPQN4oW4jSCHPJRwhUp9h2aAEniHQzpqnFvJbbG2CsQMdf99%2FHpGyxGaIfh%2B5fm2ZOEWjWrAFxLSd1ad4uJSRLfQK8IYtgZ6dSi6B%2FcxSNY3fNFqY3ATNAiMkDFfnF%2FAVzyoPH7t6VIgG9LKTpvx8kYQdR1aonq4LDotsKwjTwUhgiYmbTcr%2FE1cIHLdErbKjlQgI%2BRX%2FrXRjQGPsEOeIIlY4OT4S%2FZgr4ZR0PkqvgNxXDiJ3frnWUW%2F8eFa%2FS4iSF%2FWHbcVYPacZSw9udBJWMVfLGWw6kXpJwpblBJC05b4NPWhrQH%2BwcJOttApbVqP7P8sbRnWM6%2BABUOtXBjh1nfO4BsZ8JvZ7PVxjGF356bRvmv7B02QQ3NJ3jZdX6SmGJuOIbk2tqDOOOx09zanbcJu%2BcqI9O%2Fhiq4zAWiuyznWDvIud37Jh8MLvB0NFeyUCKIFAHNJ8PGBJeVa56TEsv%2B4PHazwDCZ1JStBjqVAjS07YXEoBoexua24awKZlieU%2F%2B2Bj%2FY9iT7Ufyb6W0uBhAxcEw1q1K728Bn3Kkygn8iD3AZjH%2BP6421sfksEovUDsZfXt7b6v0yCK0RzUhYlAiZnd0mqmFCggJnKgG4TGL8BQ5ZdqcZDWcMAAoKYzRU7MM5u8yQxblAC3exQbR01OOBFYSiDHhtikf0AjvVS%2BzCNAdyw9%2BbdPAk%2FvE6wmOxgHUZZ7pu4ofnCs1vEDmDNP8E5wYUbcahvb98P5IcinMQvQ%2FifQDo8sE3Xhq2sOXpVE3IkTdx%2FfDGI8cuMDzVzJmI1juN53tqkixUNmIzLoNSf%2F4hBU0DOb%2BZz5Gqeu1W0jxifqOPmMU932ItuY3Wy28L8cY%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Date=20240115T160635Z&X-Amz-SignedHeaders=host&X-Amz-Expires=43200&X-Amz-Credential=ASIAZXWHO2DFNXQNNMLW%2F20240115%2Feu-west-2%2Fs3%2Faws4_request&X-Amz-Signature=f79e998106566500767b3403304e323c8ce949a1a0d98bbf4224662e605b10db",
            "https://bucket-name.s3.eu-west-2.amazonaws.com/sample-data/file-2.csv?response-content-disposition=inline&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEOX%2F%2F%2F%2F%2F%2F%2F%2F%2F%2FwEaCWV1LXdlc3QtMiJGMEQCICzap9Nfx0BY9IPhrWCPEWwiR4OYW%2BxxheB%2FeorchP6QAiB5lRoXq3cjr%2BTFMvjuvGoCTaMxw9T4SzbYoUzIDMtIwyqoBAiO%2F%2F%2F%2F%2F%2F%2F%2F%2F%2F8BEAMaDDY2OTM1OTQ2MDU1NCIMAGNCpXF5mRRjsS7mKvwDcOy9VbRwNRJKZFkEdUtcXHRkKTFw1t%2FSYDaqE%2BrdEhV01bV7v5uuEJQye9zI24r%2FHJvS1dolfumoaDdjkkI6%2FtODeGo0WiZsrI8yOxCFL6WCDnEIS%2F9VJpN7MdXR9kdMhPu1crhOM52PmD0QUO6jckS1RoxXyDChrW6xB9dI7w6h%2F%2B674gHg8De%2BAvJFe45eGYqPNogwLPQN4oW4jSCHPJRwhUp9h2aAEniHQzpqnFvJbbG2CsQMdf99%2FHpGyxGaIfh%2B5fm2ZOEWjWrAFxLSd1ad4uJSRLfQK8IYtgZ6dSi6B%2FcxSNY3fNFqY3ATNAiMkDFfnF%2FAVzyoPH7t6VIgG9LKTpvx8kYQdR1aonq4LDotsKwjTwUhgiYmbTcr%2FE1cIHLdErbKjlQgI%2BRX%2FrXRjQGPsEOeIIlY4OT4S%2FZgr4ZR0PkqvgNxXDiJ3frnWUW%2F8eFa%2FS4iSF%2FWHbcVYPacZSw9udBJWMVfLGWw6kXpJwpblBJC05b4NPWhrQH%2BwcJOttApbVqP7P8sbRnWM6%2BABUOtXBjh1nfO4BsZ8JvZ7PVxjGF356bRvmv7B02QQ3NJ3jZdX6SmGJuOIbk2tqDOOOx09zanbcJu%2BcqI9O%2Fhiq4zAWiuyznWDvIud37Jh8MLvB0NFeyUCKIFAHNJ8PGBJeVa56TEsv%2B4PHazwDCZ1JStBjqVAjS07YXEoBoexua24awKZlieU%2F%2B2Bj%2FY9iT7Ufyb6W0uBhAxcEw1q1K728Bn3Kkygn8iD3AZjH%2BP6421sfksEovUDsZfXt7b6v0yCK0RzUhYlAiZnd0mqmFCggJnKgG4TGL8BQ5ZdqcZDWcMAAoKYzRU7MM5u8yQxblAC3exQbR01OOBFYSiDHhtikf0AjvVS%2BzCNAdyw9%2BbdPAk%2FvE6wmOxgHUZZ7pu4ofnCs1vEDmDNP8E5wYUbcahvb98P5IcinMQvQ%2FifQDo8sE3Xhq2sOXpVE3IkTdx%2FfDGI8cuMDzVzJmI1juN53tqkixUNmIzLoNSf%2F4hBU0DOb%2BZz5Gqeu1W0jxifqOPmMU932ItuY3Wy28L8cY%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Date=20240115T160658Z&X-Amz-SignedHeaders=host&X-Amz-Expires=43200&X-Amz-Credential=ASIAZXWHO2DFNXQNNMLW%2F20240115%2Feu-west-2%2Fs3%2Faws4_request&X-Amz-Signature=516f0a60a759b134e2dcd2fbdba1b48cf618b1e870ee9b462c6fd8facca39553",
            "https://bucket-name.s3.eu-west-2.amazonaws.com/sample-data/file-3.csv?response-content-disposition=inline&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEOX%2F%2F%2F%2F%2F%2F%2F%2F%2F%2FwEaCWV1LXdlc3QtMiJGMEQCICzap9Nfx0BY9IPhrWCPEWwiR4OYW%2BxxheB%2FeorchP6QAiB5lRoXq3cjr%2BTFMvjuvGoCTaMxw9T4SzbYoUzIDMtIwyqoBAiO%2F%2F%2F%2F%2F%2F%2F%2F%2F%2F8BEAMaDDY2OTM1OTQ2MDU1NCIMAGNCpXF5mRRjsS7mKvwDcOy9VbRwNRJKZFkEdUtcXHRkKTFw1t%2FSYDaqE%2BrdEhV01bV7v5uuEJQye9zI24r%2FHJvS1dolfumoaDdjkkI6%2FtODeGo0WiZsrI8yOxCFL6WCDnEIS%2F9VJpN7MdXR9kdMhPu1crhOM52PmD0QUO6jckS1RoxXyDChrW6xB9dI7w6h%2F%2B674gHg8De%2BAvJFe45eGYqPNogwLPQN4oW4jSCHPJRwhUp9h2aAEniHQzpqnFvJbbG2CsQMdf99%2FHpGyxGaIfh%2B5fm2ZOEWjWrAFxLSd1ad4uJSRLfQK8IYtgZ6dSi6B%2FcxSNY3fNFqY3ATNAiMkDFfnF%2FAVzyoPH7t6VIgG9LKTpvx8kYQdR1aonq4LDotsKwjTwUhgiYmbTcr%2FE1cIHLdErbKjlQgI%2BRX%2FrXRjQGPsEOeIIlY4OT4S%2FZgr4ZR0PkqvgNxXDiJ3frnWUW%2F8eFa%2FS4iSF%2FWHbcVYPacZSw9udBJWMVfLGWw6kXpJwpblBJC05b4NPWhrQH%2BwcJOttApbVqP7P8sbRnWM6%2BABUOtXBjh1nfO4BsZ8JvZ7PVxjGF356bRvmv7B02QQ3NJ3jZdX6SmGJuOIbk2tqDOOOx09zanbcJu%2BcqI9O%2Fhiq4zAWiuyznWDvIud37Jh8MLvB0NFeyUCKIFAHNJ8PGBJeVa56TEsv%2B4PHazwDCZ1JStBjqVAjS07YXEoBoexua24awKZlieU%2F%2B2Bj%2FY9iT7Ufyb6W0uBhAxcEw1q1K728Bn3Kkygn8iD3AZjH%2BP6421sfksEovUDsZfXt7b6v0yCK0RzUhYlAiZnd0mqmFCggJnKgG4TGL8BQ5ZdqcZDWcMAAoKYzRU7MM5u8yQxblAC3exQbR01OOBFYSiDHhtikf0AjvVS%2BzCNAdyw9%2BbdPAk%2FvE6wmOxgHUZZ7pu4ofnCs1vEDmDNP8E5wYUbcahvb98P5IcinMQvQ%2FifQDo8sE3Xhq2sOXpVE3IkTdx%2FfDGI8cuMDzVzJmI1juN53tqkixUNmIzLoNSf%2F4hBU0DOb%2BZz5Gqeu1W0jxifqOPmMU932ItuY3Wy28L8cY%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Date=20240115T160742Z&X-Amz-SignedHeaders=host&X-Amz-Expires=43200&X-Amz-Credential=ASIAZXWHO2DFNXQNNMLW%2F20240115%2Feu-west-2%2Fs3%2Faws4_request&X-Amz-Signature=18071e2d0cf194a27724bdd01e20ba00537d5bdeb81f43547c52cf7e654f4577",
            "https://bucket-name.s3.eu-west-2.amazonaws.com/sample-data/file-4.csv?response-content-disposition=inline&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEOX%2F%2F%2F%2F%2F%2F%2F%2F%2F%2FwEaCWV1LXdlc3QtMiJGMEQCICzap9Nfx0BY9IPhrWCPEWwiR4OYW%2BxxheB%2FeorchP6QAiB5lRoXq3cjr%2BTFMvjuvGoCTaMxw9T4SzbYoUzIDMtIwyqoBAiO%2F%2F%2F%2F%2F%2F%2F%2F%2F%2F8BEAMaDDY2OTM1OTQ2MDU1NCIMAGNCpXF5mRRjsS7mKvwDcOy9VbRwNRJKZFkEdUtcXHRkKTFw1t%2FSYDaqE%2BrdEhV01bV7v5uuEJQye9zI24r%2FHJvS1dolfumoaDdjkkI6%2FtODeGo0WiZsrI8yOxCFL6WCDnEIS%2F9VJpN7MdXR9kdMhPu1crhOM52PmD0QUO6jckS1RoxXyDChrW6xB9dI7w6h%2F%2B674gHg8De%2BAvJFe45eGYqPNogwLPQN4oW4jSCHPJRwhUp9h2aAEniHQzpqnFvJbbG2CsQMdf99%2FHpGyxGaIfh%2B5fm2ZOEWjWrAFxLSd1ad4uJSRLfQK8IYtgZ6dSi6B%2FcxSNY3fNFqY3ATNAiMkDFfnF%2FAVzyoPH7t6VIgG9LKTpvx8kYQdR1aonq4LDotsKwjTwUhgiYmbTcr%2FE1cIHLdErbKjlQgI%2BRX%2FrXRjQGPsEOeIIlY4OT4S%2FZgr4ZR0PkqvgNxXDiJ3frnWUW%2F8eFa%2FS4iSF%2FWHbcVYPacZSw9udBJWMVfLGWw6kXpJwpblBJC05b4NPWhrQH%2BwcJOttApbVqP7P8sbRnWM6%2BABUOtXBjh1nfO4BsZ8JvZ7PVxjGF356bRvmv7B02QQ3NJ3jZdX6SmGJuOIbk2tqDOOOx09zanbcJu%2BcqI9O%2Fhiq4zAWiuyznWDvIud37Jh8MLvB0NFeyUCKIFAHNJ8PGBJeVa56TEsv%2B4PHazwDCZ1JStBjqVAjS07YXEoBoexua24awKZlieU%2F%2B2Bj%2FY9iT7Ufyb6W0uBhAxcEw1q1K728Bn3Kkygn8iD3AZjH%2BP6421sfksEovUDsZfXt7b6v0yCK0RzUhYlAiZnd0mqmFCggJnKgG4TGL8BQ5ZdqcZDWcMAAoKYzRU7MM5u8yQxblAC3exQbR01OOBFYSiDHhtikf0AjvVS%2BzCNAdyw9%2BbdPAk%2FvE6wmOxgHUZZ7pu4ofnCs1vEDmDNP8E5wYUbcahvb98P5IcinMQvQ%2FifQDo8sE3Xhq2sOXpVE3IkTdx%2FfDGI8cuMDzVzJmI1juN53tqkixUNmIzLoNSf%2F4hBU0DOb%2BZz5Gqeu1W0jxifqOPmMU932ItuY3Wy28L8cY%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Date=20240115T160758Z&X-Amz-SignedHeaders=host&X-Amz-Expires=43200&X-Amz-Credential=ASIAZXWHO2DFNXQNNMLW%2F20240115%2Feu-west-2%2Fs3%2Faws4_request&X-Amz-Signature=71dbfc6d32ff820fdde98c1add38ce6ad153e450a919fe9f1368a75f8d5b3765",
        ]
        BatchURLFiles(urls).build()
        ```
        ```python
        {
            "urls": [
                "https://bucket-name.s3.eu-west-2.amazonaws.com/sample-data/file-1.csv?response-content-disposition=inline&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEOX%2F%2F%2F%2F%2F%2F%2F%2F%2F%2FwEaCWV1LXdlc3QtMiJGMEQCICzap9Nfx0BY9IPhrWCPEWwiR4OYW%2BxxheB%2FeorchP6QAiB5lRoXq3cjr%2BTFMvjuvGoCTaMxw9T4SzbYoUzIDMtIwyqoBAiO%2F%2F%2F%2F%2F%2F%2F%2F%2F%2F8BEAMaDDY2OTM1OTQ2MDU1NCIMAGNCpXF5mRRjsS7mKvwDcOy9VbRwNRJKZFkEdUtcXHRkKTFw1t%2FSYDaqE%2BrdEhV01bV7v5uuEJQye9zI24r%2FHJvS1dolfumoaDdjkkI6%2FtODeGo0WiZsrI8yOxCFL6WCDnEIS%2F9VJpN7MdXR9kdMhPu1crhOM52PmD0QUO6jckS1RoxXyDChrW6xB9dI7w6h%2F%2B674gHg8De%2BAvJFe45eGYqPNogwLPQN4oW4jSCHPJRwhUp9h2aAEniHQzpqnFvJbbG2CsQMdf99%2FHpGyxGaIfh%2B5fm2ZOEWjWrAFxLSd1ad4uJSRLfQK8IYtgZ6dSi6B%2FcxSNY3fNFqY3ATNAiMkDFfnF%2FAVzyoPH7t6VIgG9LKTpvx8kYQdR1aonq4LDotsKwjTwUhgiYmbTcr%2FE1cIHLdErbKjlQgI%2BRX%2FrXRjQGPsEOeIIlY4OT4S%2FZgr4ZR0PkqvgNxXDiJ3frnWUW%2F8eFa%2FS4iSF%2FWHbcVYPacZSw9udBJWMVfLGWw6kXpJwpblBJC05b4NPWhrQH%2BwcJOttApbVqP7P8sbRnWM6%2BABUOtXBjh1nfO4BsZ8JvZ7PVxjGF356bRvmv7B02QQ3NJ3jZdX6SmGJuOIbk2tqDOOOx09zanbcJu%2BcqI9O%2Fhiq4zAWiuyznWDvIud37Jh8MLvB0NFeyUCKIFAHNJ8PGBJeVa56TEsv%2B4PHazwDCZ1JStBjqVAjS07YXEoBoexua24awKZlieU%2F%2B2Bj%2FY9iT7Ufyb6W0uBhAxcEw1q1K728Bn3Kkygn8iD3AZjH%2BP6421sfksEovUDsZfXt7b6v0yCK0RzUhYlAiZnd0mqmFCggJnKgG4TGL8BQ5ZdqcZDWcMAAoKYzRU7MM5u8yQxblAC3exQbR01OOBFYSiDHhtikf0AjvVS%2BzCNAdyw9%2BbdPAk%2FvE6wmOxgHUZZ7pu4ofnCs1vEDmDNP8E5wYUbcahvb98P5IcinMQvQ%2FifQDo8sE3Xhq2sOXpVE3IkTdx%2FfDGI8cuMDzVzJmI1juN53tqkixUNmIzLoNSf%2F4hBU0DOb%2BZz5Gqeu1W0jxifqOPmMU932ItuY3Wy28L8cY%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Date=20240115T160635Z&X-Amz-SignedHeaders=host&X-Amz-Expires=43200&X-Amz-Credential=ASIAZXWHO2DFNXQNNMLW%2F20240115%2Feu-west-2%2Fs3%2Faws4_request&X-Amz-Signature=f79e998106566500767b3403304e323c8ce949a1a0d98bbf4224662e605b10db",
                "https://bucket-name.s3.eu-west-2.amazonaws.com/sample-data/file-2.csv?response-content-disposition=inline&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEOX%2F%2F%2F%2F%2F%2F%2F%2F%2F%2FwEaCWV1LXdlc3QtMiJGMEQCICzap9Nfx0BY9IPhrWCPEWwiR4OYW%2BxxheB%2FeorchP6QAiB5lRoXq3cjr%2BTFMvjuvGoCTaMxw9T4SzbYoUzIDMtIwyqoBAiO%2F%2F%2F%2F%2F%2F%2F%2F%2F%2F8BEAMaDDY2OTM1OTQ2MDU1NCIMAGNCpXF5mRRjsS7mKvwDcOy9VbRwNRJKZFkEdUtcXHRkKTFw1t%2FSYDaqE%2BrdEhV01bV7v5uuEJQye9zI24r%2FHJvS1dolfumoaDdjkkI6%2FtODeGo0WiZsrI8yOxCFL6WCDnEIS%2F9VJpN7MdXR9kdMhPu1crhOM52PmD0QUO6jckS1RoxXyDChrW6xB9dI7w6h%2F%2B674gHg8De%2BAvJFe45eGYqPNogwLPQN4oW4jSCHPJRwhUp9h2aAEniHQzpqnFvJbbG2CsQMdf99%2FHpGyxGaIfh%2B5fm2ZOEWjWrAFxLSd1ad4uJSRLfQK8IYtgZ6dSi6B%2FcxSNY3fNFqY3ATNAiMkDFfnF%2FAVzyoPH7t6VIgG9LKTpvx8kYQdR1aonq4LDotsKwjTwUhgiYmbTcr%2FE1cIHLdErbKjlQgI%2BRX%2FrXRjQGPsEOeIIlY4OT4S%2FZgr4ZR0PkqvgNxXDiJ3frnWUW%2F8eFa%2FS4iSF%2FWHbcVYPacZSw9udBJWMVfLGWw6kXpJwpblBJC05b4NPWhrQH%2BwcJOttApbVqP7P8sbRnWM6%2BABUOtXBjh1nfO4BsZ8JvZ7PVxjGF356bRvmv7B02QQ3NJ3jZdX6SmGJuOIbk2tqDOOOx09zanbcJu%2BcqI9O%2Fhiq4zAWiuyznWDvIud37Jh8MLvB0NFeyUCKIFAHNJ8PGBJeVa56TEsv%2B4PHazwDCZ1JStBjqVAjS07YXEoBoexua24awKZlieU%2F%2B2Bj%2FY9iT7Ufyb6W0uBhAxcEw1q1K728Bn3Kkygn8iD3AZjH%2BP6421sfksEovUDsZfXt7b6v0yCK0RzUhYlAiZnd0mqmFCggJnKgG4TGL8BQ5ZdqcZDWcMAAoKYzRU7MM5u8yQxblAC3exQbR01OOBFYSiDHhtikf0AjvVS%2BzCNAdyw9%2BbdPAk%2FvE6wmOxgHUZZ7pu4ofnCs1vEDmDNP8E5wYUbcahvb98P5IcinMQvQ%2FifQDo8sE3Xhq2sOXpVE3IkTdx%2FfDGI8cuMDzVzJmI1juN53tqkixUNmIzLoNSf%2F4hBU0DOb%2BZz5Gqeu1W0jxifqOPmMU932ItuY3Wy28L8cY%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Date=20240115T160658Z&X-Amz-SignedHeaders=host&X-Amz-Expires=43200&X-Amz-Credential=ASIAZXWHO2DFNXQNNMLW%2F20240115%2Feu-west-2%2Fs3%2Faws4_request&X-Amz-Signature=516f0a60a759b134e2dcd2fbdba1b48cf618b1e870ee9b462c6fd8facca39553",
                "https://bucket-name.s3.eu-west-2.amazonaws.com/sample-data/file-3.csv?response-content-disposition=inline&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEOX%2F%2F%2F%2F%2F%2F%2F%2F%2F%2FwEaCWV1LXdlc3QtMiJGMEQCICzap9Nfx0BY9IPhrWCPEWwiR4OYW%2BxxheB%2FeorchP6QAiB5lRoXq3cjr%2BTFMvjuvGoCTaMxw9T4SzbYoUzIDMtIwyqoBAiO%2F%2F%2F%2F%2F%2F%2F%2F%2F%2F8BEAMaDDY2OTM1OTQ2MDU1NCIMAGNCpXF5mRRjsS7mKvwDcOy9VbRwNRJKZFkEdUtcXHRkKTFw1t%2FSYDaqE%2BrdEhV01bV7v5uuEJQye9zI24r%2FHJvS1dolfumoaDdjkkI6%2FtODeGo0WiZsrI8yOxCFL6WCDnEIS%2F9VJpN7MdXR9kdMhPu1crhOM52PmD0QUO6jckS1RoxXyDChrW6xB9dI7w6h%2F%2B674gHg8De%2BAvJFe45eGYqPNogwLPQN4oW4jSCHPJRwhUp9h2aAEniHQzpqnFvJbbG2CsQMdf99%2FHpGyxGaIfh%2B5fm2ZOEWjWrAFxLSd1ad4uJSRLfQK8IYtgZ6dSi6B%2FcxSNY3fNFqY3ATNAiMkDFfnF%2FAVzyoPH7t6VIgG9LKTpvx8kYQdR1aonq4LDotsKwjTwUhgiYmbTcr%2FE1cIHLdErbKjlQgI%2BRX%2FrXRjQGPsEOeIIlY4OT4S%2FZgr4ZR0PkqvgNxXDiJ3frnWUW%2F8eFa%2FS4iSF%2FWHbcVYPacZSw9udBJWMVfLGWw6kXpJwpblBJC05b4NPWhrQH%2BwcJOttApbVqP7P8sbRnWM6%2BABUOtXBjh1nfO4BsZ8JvZ7PVxjGF356bRvmv7B02QQ3NJ3jZdX6SmGJuOIbk2tqDOOOx09zanbcJu%2BcqI9O%2Fhiq4zAWiuyznWDvIud37Jh8MLvB0NFeyUCKIFAHNJ8PGBJeVa56TEsv%2B4PHazwDCZ1JStBjqVAjS07YXEoBoexua24awKZlieU%2F%2B2Bj%2FY9iT7Ufyb6W0uBhAxcEw1q1K728Bn3Kkygn8iD3AZjH%2BP6421sfksEovUDsZfXt7b6v0yCK0RzUhYlAiZnd0mqmFCggJnKgG4TGL8BQ5ZdqcZDWcMAAoKYzRU7MM5u8yQxblAC3exQbR01OOBFYSiDHhtikf0AjvVS%2BzCNAdyw9%2BbdPAk%2FvE6wmOxgHUZZ7pu4ofnCs1vEDmDNP8E5wYUbcahvb98P5IcinMQvQ%2FifQDo8sE3Xhq2sOXpVE3IkTdx%2FfDGI8cuMDzVzJmI1juN53tqkixUNmIzLoNSf%2F4hBU0DOb%2BZz5Gqeu1W0jxifqOPmMU932ItuY3Wy28L8cY%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Date=20240115T160742Z&X-Amz-SignedHeaders=host&X-Amz-Expires=43200&X-Amz-Credential=ASIAZXWHO2DFNXQNNMLW%2F20240115%2Feu-west-2%2Fs3%2Faws4_request&X-Amz-Signature=18071e2d0cf194a27724bdd01e20ba00537d5bdeb81f43547c52cf7e654f4577",
                "https://bucket-name.s3.eu-west-2.amazonaws.com/sample-data/file-4.csv?response-content-disposition=inline&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEOX%2F%2F%2F%2F%2F%2F%2F%2F%2F%2FwEaCWV1LXdlc3QtMiJGMEQCICzap9Nfx0BY9IPhrWCPEWwiR4OYW%2BxxheB%2FeorchP6QAiB5lRoXq3cjr%2BTFMvjuvGoCTaMxw9T4SzbYoUzIDMtIwyqoBAiO%2F%2F%2F%2F%2F%2F%2F%2F%2F%2F8BEAMaDDY2OTM1OTQ2MDU1NCIMAGNCpXF5mRRjsS7mKvwDcOy9VbRwNRJKZFkEdUtcXHRkKTFw1t%2FSYDaqE%2BrdEhV01bV7v5uuEJQye9zI24r%2FHJvS1dolfumoaDdjkkI6%2FtODeGo0WiZsrI8yOxCFL6WCDnEIS%2F9VJpN7MdXR9kdMhPu1crhOM52PmD0QUO6jckS1RoxXyDChrW6xB9dI7w6h%2F%2B674gHg8De%2BAvJFe45eGYqPNogwLPQN4oW4jSCHPJRwhUp9h2aAEniHQzpqnFvJbbG2CsQMdf99%2FHpGyxGaIfh%2B5fm2ZOEWjWrAFxLSd1ad4uJSRLfQK8IYtgZ6dSi6B%2FcxSNY3fNFqY3ATNAiMkDFfnF%2FAVzyoPH7t6VIgG9LKTpvx8kYQdR1aonq4LDotsKwjTwUhgiYmbTcr%2FE1cIHLdErbKjlQgI%2BRX%2FrXRjQGPsEOeIIlY4OT4S%2FZgr4ZR0PkqvgNxXDiJ3frnWUW%2F8eFa%2FS4iSF%2FWHbcVYPacZSw9udBJWMVfLGWw6kXpJwpblBJC05b4NPWhrQH%2BwcJOttApbVqP7P8sbRnWM6%2BABUOtXBjh1nfO4BsZ8JvZ7PVxjGF356bRvmv7B02QQ3NJ3jZdX6SmGJuOIbk2tqDOOOx09zanbcJu%2BcqI9O%2Fhiq4zAWiuyznWDvIud37Jh8MLvB0NFeyUCKIFAHNJ8PGBJeVa56TEsv%2B4PHazwDCZ1JStBjqVAjS07YXEoBoexua24awKZlieU%2F%2B2Bj%2FY9iT7Ufyb6W0uBhAxcEw1q1K728Bn3Kkygn8iD3AZjH%2BP6421sfksEovUDsZfXt7b6v0yCK0RzUhYlAiZnd0mqmFCggJnKgG4TGL8BQ5ZdqcZDWcMAAoKYzRU7MM5u8yQxblAC3exQbR01OOBFYSiDHhtikf0AjvVS%2BzCNAdyw9%2BbdPAk%2FvE6wmOxgHUZZ7pu4ofnCs1vEDmDNP8E5wYUbcahvb98P5IcinMQvQ%2FifQDo8sE3Xhq2sOXpVE3IkTdx%2FfDGI8cuMDzVzJmI1juN53tqkixUNmIzLoNSf%2F4hBU0DOb%2BZz5Gqeu1W0jxifqOPmMU932ItuY3Wy28L8cY%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Date=20240115T160758Z&X-Amz-SignedHeaders=host&X-Amz-Expires=43200&X-Amz-Credential=ASIAZXWHO2DFNXQNNMLW%2F20240115%2Feu-west-2%2Fs3%2Faws4_request&X-Amz-Signature=71dbfc6d32ff820fdde98c1add38ce6ad153e450a919fe9f1368a75f8d5b3765",
            ]
        }
        ```
        """
        self.urls = urls

    def build(self) -> dict[str, Any]:
        """Build a BatchURLFiles document from given inputs."""
        return {"urls": self.urls}


class BatchS3Files:
    """A constructor class for building an BatchS3Files document. Indicates that this import batch should source files from S3."""

    def __init__(
        self, bucket: str, credentials: AWSCredentials, glob: str, prefix: str | None
    ) -> None:
        """Initilaize class and stores values.

        Examples
        --------
        ```python
        from dfi.services.ingest import AWSCredentials, BatchS3Files

        role_arn = "arn:aws:s3:::example_bucket/example_key"
        credentials = AWSCredentials(role_arn)
        bucket = "s3://bucket-name"
        glob = "*.csv"
        prefix = "sample-data"

        BatchS3Files(bucket, credentials, glob, prefix).build()
        ```
        ```python
        {
            "s3": {
                "bucket": "s3://bucket-name",
                "credentials": {
                    "RoleArn": "arn:aws:s3:::example_bucket/example_key"
                },
                "glob": "*.csv",
                "prefix": "sample-data",
            }
        }
        ```

        Parameters
        ----------
        bucket:
            The S3 bucket to download files from. This can be an s3:// format url, or a full URL.
        credentials:
            Parameters which point to an AWS role that the importer will assume to scan and download files.
            The `AWSCredentials` builder can help construct this document.
        glob:
            A filename descriptor which indicates what format files to download.

            - example: `**/*.csv`
        prefix:
            Only download files in the S3 bucket which start with this prefix.
        """
        self.bucket = bucket
        self.credentials = credentials
        self.glob = glob
        self.prefix = prefix

    def build(self) -> dict[str, Any]:
        """Build a BatchS3Files document from given inputs."""
        s3_source_descriptor = {
            "bucket": self.bucket,
            "credentials": self.credentials.build(),
            "glob": self.glob,
        }

        if self.prefix:
            s3_source_descriptor.update({"prefix": self.prefix})

        return {"s3": s3_source_descriptor}


class CSVFormat:
    """A constructor class for building an CSVFormat document.

    An object indicating how to map columns in a CSV file into points to be imported to the Data Flow Index
    """

    def __init__(
        self,
        entity_id: int,
        timestamp: int,
        longitude: int,
        latitude: int,
        altitude: int | None = None,
        metadata_id: int | None = None,
        **kwargs: int,
    ) -> None:
        """Initialize class and store values.

        Examples
        --------
        ```python
        from dfi.services.ingest import CSVFormat

        # required fields
        entity_id = 0
        timestamp = 1
        longitude = 2
        latitude = 3

        # extra fields
        ipv4 = 4
        age = 5
        home_ipv4 = 6
        route_id = 7
        credit_card_provider = 8
        transportation_mode = 9

        CSVFormat(
            entity_id=entity_id,
            timestamp=timestamp,
            longitude=longitude,
            latitude=latitude,
            ipv4=ipv4,
            age=age,
            home_ipv4=home_ipv4,
            route_id=route_id,
            credit_card_provider=credit_card_provider,
            transportation_mode=transportation_mode,
        ).build()
        ```
        ```python
        {
            "csv": {
                "entityId": 0,
                "timestamp": 1,
                "longitude": 2,
                "latitude": 3,
                "ipv4": 4,
                "age": 5,
                "home_ipv4": 6,
                "route_id": 7,
                "credit_card_provider": 8,
                "transportation_mode": 9,
            }
        }
        ```

        Parameters
        ----------
        entity_id:
            The column number to use as the `entityId`.
        timestamp:
            The column number to use as the `timestamp`.
        longitude:
            The column number to use as the `longitude`.
        latitude:
            The column number to use as the `latitude`.
        kwargs:
            The column number to use for the given keyword argument.
        """
        self.entity_id = entity_id
        self.timestamp = timestamp
        self.longitude = longitude
        self.latitude = latitude
        self.altitude = altitude
        self.metadata_id = metadata_id
        self.kwargs = kwargs

    def build(self) -> dict[str, Any]:
        """Build a CSVFormat document from given inputs."""
        csv_format: dict[str, Any] = {
            "entityId": self.entity_id,
            "timestamp": self.timestamp,
            "longitude": self.longitude,
            "latitude": self.latitude,
        }
        if self.altitude:
            csv_format.update({"altitude": self.altitude})
        if self.altitude:
            csv_format.update({"metadataId": self.metadata_id})
        csv_format.update(self.kwargs)

        return {"csv": csv_format}


class S3URLPresigner:
    """For creating presigned URLs to objects in AWS S3."""

    def __init__(self, bucket: str, region: str, profile: str | None) -> None:
        """Initialize and store paramters and create a connection to AWS S3.

        Parameters
        ----------
        bucket:
            S3 bucket.
        region:
            S3 region.
        profile:
            S3 profile to use when accessing resources.
        """
        self.bucket = bucket
        self.region = region
        self.profile = profile

        boto3.setup_default_session(profile_name=self.profile)
        config = Config(region_name=self.region, signature_version="s3v4")

        self.s3_client = boto3.client("s3", config=config)
        self.s3_resource = boto3.resource("s3", config=config)

    def find_files(
        self, prefix: str, suffix: str, sort: bool = True, verbose: bool = False
    ) -> list[str]:
        """Globs for files in S3 Bucket/Prefix ending with suffix.  Will sort files by name using natural sort.

        Parameters
        ----------
        prefix:
            S3 prefix to search in.
        suffix:
            Will search for files ending with `suffix`.  Example: suffix=`".csv"` will find files inding in `".csv"`.
        sort:
            Whether to sort the files.  Sort is done via natural sort.
        verbose:
            If `True`, will show a progress bar.

        Returns
        -------
        files:
            A List of files found in the S3 bucket.

        Examples
        --------
        The example below will find all CSV files (files ending with `".csv"`) in `"s3://datasets/dataset-1"`.
        ```python
        from dfi.services.ingest import S3URLPresigner

        bucket = "datasets"
        region = "eu-west-2"
        profile = "default"

        s3_presigner = S3URLPresigner(bucket, region, profile)

        files = s3_presigner.list_files(object_key, expiration=720)
        ```
        """
        bucket = self.s3_resource.Bucket(self.bucket)

        files = []
        for s3_object in tqdm(
            bucket.objects.filter(Prefix=prefix),
            desc=f"Globbing files ending with '{suffix}' in '{self.bucket}/{prefix}'",
            disable=not verbose,
        ):
            if s3_object.key.endswith(suffix):
                files.append(s3_object.key)

        if sort:
            files = natsort.natsorted(files)
        return files

    def generate_presigned_url(
        self,
        object_key: str,
        expiration: int,
    ) -> str:
        """Create a presigned URL for the key.

        Parameters
        ----------
        object_key:
            S3 object key.
        expiration:
            Time in minutes for the presigned URL to remain valid.  min=1, max=720

        Returns
        -------
        presigned URL
            URL will be valid for the set `expiration` time.

        Examples
        --------
        The example below will generate a presigned URL for `"s3://datasets/dataset-1/file-1.csv"`.
        ```python
        from dfi.services.ingest import S3URLPresigner

        bucket = "datasets"
        region = "eu-west-2"
        profile = "default"
        s3_presigner = S3URLPresigner(bucket, region, profile)

        object_key = "dataset-1/file-1.csv"
        presigned_url = s3_presigner.generate_presigned_url(
            object_key, expiration=720
        )
        ```
        """
        presigned_urls: str = self.s3_client.generate_presigned_url(
            "get_object",
            Params={"Bucket": self.bucket, "Key": object_key},
            ExpiresIn=expiration,
        )
        return presigned_urls

    def generate_presigned_urls(
        self,
        prefix: str,
        suffix: str,
        expiration: int,
        sort: bool = True,
        verbose: bool = False,
    ) -> list[str]:
        """Create presigned URLs matching the glob pattern in the bucket.

        Parameters
        ----------
        prefix:
            S3 object prefix
        suffix:
            Suffix patten to search during glob
        expiration:
            Time in minutes for the presigned URL to remain valid.  min=1, max=720
        sort:
            Whether to sort the files.  Sort is done via natural sort.
        verbose:
            If true, will show a progress bar.

        Returns
        -------
        presigned URLs
            A list of presigned URLs. If error, returns `None`.

        Examples
        --------
        The example below will generate presigned URLs for all CSV files (files ending with `".csv"`)
            in `s3://datasets/dataset-1`.
        ```python
        from dfi.services.ingest import S3URLPresigner

        bucket = "datasets"
        region = "eu-west-2"
        profile = "default"
        prefix = "dataset-1"

        s3_presigner = S3URLPresigner(bucket, region, profile)
        presigned_urls = s3_presigner.generate_presigned_urls(
            prefix, ".csv", expiration=720, sort=True, verbose=False
        )
        ```
        """
        files = self.find_files(prefix, suffix, sort=sort, verbose=verbose)

        presigned_urls = []
        for file in tqdm(files, desc="Presigning URLs", disable=not verbose):
            url = self.s3_client.generate_presigned_url(
                "get_object",
                Params={"Bucket": self.bucket, "Key": file},
                ExpiresIn=expiration,
            )
            presigned_urls.append(url)

        return presigned_urls


class Ingest:
    """Class responsible for handling data ingests."""

    def __init__(self, conn: Connect) -> None:
        self.conn = conn

    def __repr__(self) -> str:
        """Class representation."""
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        """Class string formatting."""
        return f"""Instance of dfi.{self.__class__.__name__} composed with: {self.conn!s}"""

    def get_aws_trust_policy(self) -> dict[Any, Any]:
        """Retrieve the AWS Trust Policy for the Import API.

        When importing data from S3, we need to assume a role in your AWS account to be able to scan and
        download data. This requires adding a trust policy to be added to your role. You can get a copy
        of this trust policy here.

        ??? info "Endpoint"
            [GET /v1/import/awsTrustPolicy](https://api.prod.generalsystem.com/docs/api#/Data%20Import%20(v1)/get_v1_import_awsTrustPolicy)

        Returns
        -------
        trust policy
            An AWS Trust Policy for the General System API.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)
        dfi.ingest.get_aws_trust_policy()
        ```
        """
        with self.conn.api_get("v1/import/awsTrustPolicy", stream=False) as response:
            response.raise_for_status()
            aws_trust_policy: dict[Any, Any] = response.json()
            return aws_trust_policy

    def put_batch(
        self,
        dataset_id: str,
        source: BatchURLFiles | BatchS3Files,
        file_format: CSVFormat,
        dry_run: bool | int = False,
    ) -> dict[str, Any]:
        """Ingest files into a GS Platform dataset.

        Ingest can be from either:

        1. a list of URLs - these can be pre-signed AWS S3 URLs, or any URL with open access to a CSV behind it.
        2. an AWS S3 bucket the user has obtained access through an AWS Trust Policy.

        ??? info "Endpoint"
            [PUT /v1/import/batch](https://api.prod.generalsystem.com/docs/api#/Data%20Import%20(v1)/put_v1_import_batch)

        ??? tip
            See the Ingest User Guide for full examples.

        Parameters
        ----------
        dataset_id:
            the dataset to ingest data into.
        source:
            This can be either a batch of URLs pointing to files or an S3 bucket + prefix with files
            to be ingested.  Refer to `BatchURLFiles` or `BatchS3Files` for building the dictionary.
        file_format:
            A dictionary indicating how to map columns in a CSV file into points to be imported to
            the Data Flow Index.  Refer to `CSVFormat` for building the dictionary.
        dry_run:
            Set to `True` to perform a dry run that checks the first 100 rows of each file.
            Set to an integer (e.g. `1_000`) to perform a dry run that checks the first 1,000 rows of each file.
            No data will be imported into the Data Flow Index.

        Returns
        -------
        ingest info
            information about the ingest.  If `dry_run=True` a report of the run will be returned.
        """
        if isinstance(source, BatchS3Files):
            raise NotImplementedError(
                "'BatchS3Files` stubs are present but not yet implemented.  Instead use `BatchURLFiles`."
            )

        payload = {
            "datasetId": dataset_id,
            "source": source.build(),
            "format": file_format.build(),
        }

        # requests will serialize True / False into "True" / "False", which is not a JSON boolean
        params = {"dryRun": json.dumps(dry_run)}

        with self.conn.api_put(
            "v1/import/batch", params=params, json=payload, stream=False
        ) as response:
            response.raise_for_status()
            ingest_info: dict[str, Any] = response.json()
            return ingest_info

    def get_batch_info(
        self,
        import_batch_id: str,
    ) -> dict[str, Any]:
        """Retrieve information about a previously created import batch.

        ??? info "Endpoint"
            [GET /v1/import/batch/{id}](https://api.prod.generalsystem.com/docs/api#/Data%20Import%20(v1)/get_v1_import_batch__importBatchId_)

        Parameters
        ----------
        import_batch_id:
            the batch id of the import.

        Returns
        -------
        batch info
            information about the imported batch.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)
        dfi.ingest.get_batch_info(<import_batch_id>)
        ```
        """
        with self.conn.api_get(
            f"v1/import/batch/{import_batch_id}", stream=False
        ) as response:
            response.raise_for_status()
            batch_info: dict[str, Any] = response.json()
            return batch_info

    def update_batch_status(self, import_batch_id: str, status: str) -> dict[str, str]:
        """Update information about a previously created import batch.

        ??? info "Endpoint"
            [PATCH /v1/import/batch/{id}](https://api.prod.generalsystem.com/docs/api#/Data%20Import%20(v1)/patch_v1_import_batch__importBatchId_)

        ??? note "Aborting an import"
            This route can be used to abort an in-progress batch by setting the `status` field to `aborted`.

        Parameters
        ----------
        import_batch_id:
            the batch id of the import.
        status:
            the status to update the import to.  One of [`aborted`]

        Returns
        -------
        status
            the updated status of the batch

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)

        status = "aborted"
        dfi.ingest.update_batch_status(<import_batch_id>, status)
        ```
        """
        body = {"status": status}
        with self.conn.api_patch(
            f"v1/import/batch/{import_batch_id}", json=body, stream=False
        ) as response:
            response.raise_for_status()
            batch_status: dict[str, str] = response.json()
            return batch_status

    def get_batch_status(self, import_batch_id: str) -> list[dict[str, Any]]:
        """Retrieve a chronological series of status updates about an import batch.

        ??? info "Endpoint"
            [GET /v1/import/batch/{id}/status](https://api.prod.generalsystem.com/docs/api#/Data%20Import%20(v1)/get_v1_import_batch__importBatchId__status)

        Parameters
        ----------
        import_batch_id:
            the batch id of the import.

        Returns
        -------
        statuses
            a list of status updates for the batch.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)
        dfi.ingest.get_batch_status(<import_batch_id>)
        ```
        """
        with self.conn.api_get(
            f"v1/import/batch/{import_batch_id}/status", stream=False
        ) as response:
            response.raise_for_status()
            batch_status: list[dict[str, Any]] = response.json()
            return batch_status



"""BBox model definiton."""

from collections.abc import Sequence

from typing_extensions import Self

from dfi.errors import BBoxLatitudeMismatchError, BBoxLongitudeMismatchError, BBoxUndefinedError, BBoxValueError
from dfi.models.filters.geometry import Point

BBOX_LENGTH = 4


class BBox:
    """A 2D BBox as defined in [GeoJSON BBox](https://datatracker.ietf.org/doc/html/rfc7946#section-5)."""

    _min_lon: float
    _min_lat: float
    _max_lon: float
    _max_lat: float

    def __repr__(self) -> str:
        """Class representation."""
        return f"BBox([{self._min_lon}, {self._min_lat}, {self._max_lon}, {self._max_lat}])"

    def __str__(self) -> str:
        """Class string formatting."""
        return f"BBox([{self._min_lon}, {self._min_lat}, {self._max_lon}, {self._max_lat}])"

    @property
    def min_lon(self) -> float:
        """The min_lon property."""
        return self._min_lon

    @property
    def min_lat(self) -> float:
        """The min_lat property."""
        return self._min_lat

    @property
    def max_lon(self) -> float:
        """The max_lon property."""
        return self._max_lon

    @property
    def max_lat(self) -> float:
        """The max_lat property."""
        return self._max_lat

    def from_corners(self, min_lon: float, min_lat: float, max_lon: float, max_lat: float) -> Self:
        """Create a BBox from specified longitude & latitude values.

        Parameters
        ----------
        min_lon:
            Minimum bound of the longitude.
        min_lat:
            Minimum bound of the latitude.
        max_lon:
            Maximum bound of the longitude.
        max_lat:
            Maximum bound of the latitude.

        Returns
        -------
        BBox

        Examples
        --------
        ### BBox From Corners
        ```python
        min_lon = 0.0
        min_lat = 0.0
        max_lon = 1.0
        max_lat = 1.0

        BBox().from_corners(min_lon, min_lat, max_lon, max_lat)
        ```
        ```python
        BBox([0.0, 0.0, 1.0, 1.0])
        ```
        """
        self._min_lon = min_lon
        self._min_lat = min_lat
        self._max_lon = max_lon
        self._max_lat = max_lat

        return self.validate()

    def from_list(self, bounds: list[float]) -> Self:
        """Create a BBox from a GeoJSON BBox.

        Parameters
        ----------
        bounds:
            a list of 4 floats representing [min_lon, min_lat, max_lon, max_lat].

        Returns
        -------
        BBox

        Raises
        ------
        BBoxValueError
        BBoxLongitudeMismatchError
        BBoxLatitudeMismatchError
        TypeError

        Examples
        --------
        ### BBox From List
        ```python
        BBox().from_list([0.0, 0.0, 1.0, 1.0])
        ```
        ```python
        BBox([0.0, 0.0, 1.0, 1.0])
        ```
        """
        match bounds:
            case list():
                if len(bounds) != BBOX_LENGTH:
                    raise BBoxValueError(f"BBox is defined from 4 values, {len(bounds)} given.")
                self._min_lon = bounds[0]
                self._min_lat = bounds[1]
                self._max_lon = bounds[2]
                self._max_lat = bounds[3]
            case _:
                raise TypeError("'bounds' is not of type list[float].")

        return self.validate()

    def validate(self) -> Self:
        """Validate the BBox.

        Returns
        -------
        BBox

        Raises
        ------
        BBoxLongitudeMismatchError
        BBoxLatitudeMismatchError
        """
        if not (
            hasattr(self, "_min_lon")
            or hasattr(self, "_min_lat")
            or hasattr(self, "_max_lon")
            or hasattr(self, "_max_lat")
        ):
            raise BBoxUndefinedError

        Point(self._min_lon, self._min_lat).validate()
        Point(self._max_lon, self._max_lat).validate()

        if self._min_lon >= self._max_lon:
            raise BBoxLongitudeMismatchError(f"min_lon ({self._min_lon}) is >= max_lon ({self._max_lon})")
        if self._min_lat >= self._max_lat:
            raise BBoxLatitudeMismatchError(f"min_lat ({self._min_lat}) is >= max_lat ({self._max_lat})")

        return self

    def build(self) -> dict[str, str | Sequence[float]]:
        """Validate the BBox is defined and format the BBox for use in Query Document."""
        self.validate()
        return {"type": "BoundingBox", "bounds": (self._min_lon, self._min_lat, self._max_lon, self._max_lat)}



"""A Point or vertice of a Polygon."""

from typing_extensions import Self

from dfi.errors import AltitudeOutOfBoundsError, LatitudeOutOfBoundsError, LongitudeOutOfBoundsError

LONGITUDE_MAX = 180.0
LONGITUDE_MIN = -180.0
LATITUDE_MIN = -90.0
LATITUDE_MAX = 90.0
ALTITUDE_MIN = -1.7976931348623157e308
ALTITUDE_MAX = 1.7976931348623157e308


class Point:
    """A vertice of a Polygon."""

    def __init__(self, lon: float, lat: float) -> None:
        """Create a Point from coordinates and checks bounds.

        Parameters
        ----------
        lon:
            longitude
        lat:
            latitude

        Raises
        ------
        LongitudeOutOfBoundsError
        LatitudeOutOfBoundsError

        Examples
        --------
        ```python
        Point(0.0, 1.0)
        ```
        ```python
        Point(0.0, 1.0)
        ```
        """
        self._lon = lon
        self._lat = lat
        self.validate()

    def __repr__(self) -> str:
        """Class representation."""
        return f"Point({self._lon}, {self._lat})"

    def __str__(self) -> str:
        """Class string formatting."""
        return f"Point({self._lon}, {self._lat})"

    def __eq__(self, other: object) -> bool:
        """Equality check."""
        if not isinstance(other, Point):
            return NotImplemented
        return self.__dict__ == other.__dict__

    @property
    def lon(self) -> float:
        """The longitude property."""
        return self._lon

    @property
    def lat(self) -> float:
        """The latitude property."""
        return self._lat

    @staticmethod
    def _validate_latitude(lat: float) -> None:
        """Validate the bounds of the latitude.

        Raises
        ------
        LatitudeOutOfBoundsError
        """
        if not LATITUDE_MIN < lat < LATITUDE_MAX:
            raise LatitudeOutOfBoundsError(f"Latitude value '{lat}' not within ({LATITUDE_MIN}, {LATITUDE_MAX})")

    @staticmethod
    def _validate_longitude(lon: float) -> None:
        """Validate the bounds of the latitude.

        Raises
        ------
        LongitudeOutOfBoundsError
        """
        if not LONGITUDE_MIN < lon < LONGITUDE_MAX:
            raise LongitudeOutOfBoundsError(f"Longitude value '{lon}' not within ({LONGITUDE_MIN}, {LONGITUDE_MAX})")

    @staticmethod
    def _validate_altitude(alt: float) -> None:
        """Validate the bounds of the latitude.

        Raises
        ------
        AltitudeOutOfBoundsError
        """
        if not ALTITUDE_MIN < alt < ALTITUDE_MAX:
            raise AltitudeOutOfBoundsError(f"Altitude value '{alt}' not within ({ALTITUDE_MIN}, {ALTITUDE_MAX})")

    def validate(self) -> Self:
        """Validate the Point.

        Returns
        -------
        Point

        Raises
        ------
        LongitudeOutOfBoundsError
        LatitudeOutOfBoundsError
        """
        self._validate_longitude(self._lon)
        self._validate_latitude(self._lat)
        return self

    def build(self) -> tuple[float, float]:
        """Return values formatted for the Query Document."""
        return (self._lon, self._lat)



"""Class to connect to the DFI server."""

import logging

import requests

from dfi import validate

_logger = logging.getLogger(__name__)


class Connect:
    """Class instantiating the connectors to the DFI API.

    Parameters
    ----------
    api_token: token provided by generalsystem.com to access the running DFI environments.
    base_url: Base url where the service is located
    query_timeout: Time after an unresponsive query will be dropped.
    progress_bar: Visualise a progress bar if True (slows down the execution, typically used for demos and tests).

    Examples
    --------
    ```python
    connection = dfi.Connect("<token>", "<base_url>")
    ```
    """

    def __init__(
        self,
        api_token: str,
        base_url: str | None = None,
        query_timeout: int | None = 60,
        progress_bar: bool | None = False,
    ) -> None:
        self.api_token = api_token
        self.base_url = base_url
        self.query_timeout = query_timeout
        self.streaming_headers = {
            "Authorization": f"Bearer {api_token}",
            "Accept": "text/event-stream",
        }
        self.synchronous_headers = {
            "Authorization": f"Bearer {api_token}",
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        self.progress_bar = progress_bar

    def __repr__(self) -> str:
        """Class representation."""
        return f"{self.__class__.__name__}(api_token=<***>, base_url={self.base_url}, query_timeout={self.query_timeout}, progress_bar={self.progress_bar})"

    def __str__(self) -> str:
        """Class string formatting."""
        return f"{self.__class__.__name__}(api_token=<***>, base_url={self.base_url}, query_timeout={self.query_timeout}, progress_bar={self.progress_bar})"

    def api_get(
        self,
        endpoint: str,
        stream: bool = True,
        params: dict | None = None,  # type: ignore
    ) -> requests.models.Response:
        """Wrap requests.get method.

        Parameters
        ----------
        endpoint:
            The endpoint of the URL.  Will be added as a suffix to the base_url.
        stream:
            Whether to use streaming headers or synchronous headers.
        params:
            Dictionary, list of tuples or bytes to send in the query string for the request.
        """
        headers = self.streaming_headers if stream else self.synchronous_headers
        url = f"{self.base_url}/{endpoint}"
        _logger.debug(
            dict(
                url=url,
                headers=headers,
                stream=stream,
                params=params,
                timeout=self.query_timeout,
            )
        )
        response = requests.get(
            url,
            headers=headers,
            stream=stream,
            params=params,
            timeout=self.query_timeout,
        )

        validate.response(response, url, headers, params)
        return response

    def api_post(
        self,
        endpoint: str,
        stream: bool = True,
        params: dict | None = None,  # type: ignore
        json: dict | list | None = None,  # type: ignore
        data: dict | None = None,  # type: ignore
    ) -> requests.models.Response:
        """Wrap requests.post method.

        Parameters
        ----------
        endpoint:
            The endpoint of the URL.  Will be added as a suffix to the base_url.
        stream:
            Whether to use streaming headers or synchronous headers.
        params:
            Dictionary, list of tuples or bytes to send in the query string for the request.
        data:
            Dictionary, list of tuples, bytes, or file-like object to send in the body of the request
        json:
            A JSON serializable Python object to send in the body of the request.  Will set the
            "Content-Type: application/json" in the header.
        """
        headers = self.streaming_headers if stream else self.synchronous_headers
        url = f"{self.base_url}/{endpoint}"

        response = requests.post(
            url,
            headers=headers,
            json=json,
            data=data,
            stream=stream,
            params=params,
            timeout=self.query_timeout,
        )
        validate.response(response, url, headers, params, json)
        return response

    def api_put(
        self,
        endpoint: str,
        stream: bool = True,
        params: dict | None = None,  # type: ignore
        json: dict | None = None,  # type: ignore
        data: dict | None = None,  # type: ignore
    ) -> requests.models.Response:
        """Wrap requests.put method.

        Parameters
        ----------
        endpoint:
            The endpoint of the URL.  Will be added as a suffix to the base_url.
        stream:
            Whether to use streaming headers or synchronous headers.
        params:
            Dictionary, list of tuples or bytes to send in the query string for the request.
        data:
            Dictionary, list of tuples, bytes, or file-like object to send in the body of the request
        json:
            A JSON serializable Python object to send in the body of the request.  Will set the
            "Content-Type: application/json" in the header.
        """
        headers = self.streaming_headers if stream else self.synchronous_headers
        url = f"{self.base_url}/{endpoint}"

        response = requests.put(
            url,
            headers=headers,
            json=json,
            data=data,
            stream=stream,
            params=params,
            timeout=self.query_timeout,
        )
        validate.response(response, url, headers, params, json)
        return response

    def api_delete(
        self,
        endpoint: str,
        stream: bool = True,
        params: dict | None = None,  # type: ignore
        json: dict | list | None = None,  # type: ignore
        data: dict | None = None,  # type: ignore
    ) -> requests.models.Response:
        """Wrap requests.delete method.

        Parameters
        ----------
        endpoint:
            The endpoint of the URL.  Will be added as a suffix to the base_url.
        stream:
            Whether to use streaming headers or synchronous headers.
        params:
            Dictionary, list of tuples or bytes to send in the query string for the request.
        data:
            Dictionary, list of tuples, bytes, or file-like object to send in the body of the request
        json:
            A JSON serializable Python object to send in the body of the request.  Will set the
            "Content-Type: application/json" in the header.
        """
        headers = self.streaming_headers if stream else self.synchronous_headers

        url = f"{self.base_url}/{endpoint}"
        response = requests.delete(
            url,
            headers=headers,
            json=json,
            data=data,
            stream=stream,
            params=params,
            timeout=self.query_timeout,
        )
        validate.response(response, url, headers, params, json)
        return response

    def api_patch(
        self,
        endpoint: str,
        stream: bool = True,
        params: dict | None = None,  # type: ignore
        json: dict | None = None,  # type: ignore
        data: dict | None = None,  # type: ignore
    ) -> requests.models.Response:
        """Wrap requests.patch method.

        Parameters
        ----------
        endpoint:
            The endpoint of the URL.  Will be added as a suffix to the base_url.
        stream:
            Whether to use streaming headers or synchronous headers.
        params:
            Dictionary, list of tuples or bytes to send in the query string for the request.
        data:
            Dictionary, list of tuples, bytes, or file-like object to send in the body of the request
        json:
            A JSON serializable Python object to send in the body of the request.  Will set the
            "Content-Type: application/json" in the header.
        """
        headers = self.streaming_headers if stream else self.synchronous_headers
        url = f"{self.base_url}/{endpoint}"

        response = requests.patch(
            url,
            headers=headers,
            json=json,
            data=data,
            stream=stream,
            params=params,
            timeout=self.query_timeout,
        )
        validate.response(response, url, headers, params, json)
        return response



"""Manages queries to the Datasets API.  Allows users to find, create, update, and delete datasets, dataset schemas, and dataset permissions.

Tests are dependent on other test.  The order in which tests are run matters.
"""

from datetime import datetime
from typing import Any

import pyarrow as pa
import requests
from pyarrow import feather

from dfi.connect import Connect
from dfi.errors import DFIResponseError  # noqa: F401 (import to allow internal linking)


class Datasets:
    """Class responsible for handling datasets and dataset schemas."""

    def __init__(self, conn: Connect) -> None:
        self.conn = conn

    def __repr__(self) -> str:
        """Class representation."""
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        """Class string formatting."""
        return f"""Instance of dfi.{self.__class__.__name__} composed with: {self.conn!s}"""

    def create(self, dataset: dict[str, Any]) -> dict[str, Any]:
        """Create an empty dataset.

        ??? info "Endpoint"
            [POST /v1/datasets](https://api.prod.generalsystem.com/docs/api#/Dataset%20Management%20(v1)/post_v1_datasets)

        ??? tip "Admin Request"
            You need to be an admin for this request.

        Parameters
        ----------
        dataset:
            the dataset definition to be created.

        Returns
        -------
        dataset:
            The `dataset_id` is returned under the `id` key.

        Examples
        --------
        ```python
        from dfi import Client

        dataset = {
            "name": "test-0",
            "description": None,
            "active": True,
            "tags": {},
            "type": "managed",
            "model": "point",
            "dataDescription": {
                "metadataSchema": {
                    "plantHeight": {
                        "type": "number",
                        "nullable": False,
                        "signed": False,
                    },
                    "plantIPv4": {"type": "ip", "nullable": True},
                    "plantCultivar": {
                        "type": "enum",
                        "nullable": True,
                        "values": [
                            "broccoli",
                            "brocollini",
                            "brussel sprouts",
                            "cabbage",
                            "cauliflower",
                            "collards",
                        ],
                    },
                },
                "boundingBox": [-180.0, -90.0, 180.0, 90.0],
                "minDatetime": "2021-01-01T00:00:00.000Z",
                "maxDatetime": "2022-01-01T00:00:00.000Z",
            },
            "source": {"s3SourceUrl": "s3://test-bucket/dataset-0"},
            "pipeline": {"curationConfiguration": {}},
            "pii": {"keepPii": False, "piiFields": []},
            "storage": {
                "dataStoreType": "dfi",
                "dataStoreConnectionDetails": {
                    "host": "0.0.0.0",
                    "port": "1234",
                    "queryTimeout": "3600000",
                },
            },
            "destination": {"dataStoreRetentionLength": 0},
            "permissions": [{"type": "reader", "scope": "all"}],
        }

        dfi = Client(token, url)
        dfi.datasets.create(dataset)
        ```

        """
        with self.conn.api_post("v1/datasets", stream=False, json=dataset) as response:
            response.raise_for_status()
            dataset_info: dict[str, Any] = response.json()
            return dataset_info

    def find(
        self,
        name: str | None = None,
        before: datetime | None = None,
        limit: int | None = None,
    ) -> list[dict[str, Any]]:
        """Find datasets.

        ??? info "Endpoint"
            [GET /v1/datasets](https://api.prod.generalsystem.com/docs/api#/Dataset%20Management%20(v1)/get_v1_datasets)

        Parameters
        ----------
        name:
            retrieves all datasets with `name`.
        before:
            retrieves all datasets created before `before`.
        limit:
            only return `limit` number of datasets.

        Returns
        -------
        dataset:
            a list of dataset definitions.

        Examples
        --------
        ```python
        from datetime import datetime
        from dfi import Client

        dfi = Client(token, url)

        time = datetime.fromisoformat("2023-03-31T12:50:00Z")
        dfi.datasets.find(name="London", before=time, limit=10)
        ```

        """
        params = {
            "name": name,
            "before": before.isoformat() if before else None,
            "limit": limit,
        }

        with self.conn.api_get("v1/datasets", stream=False, params=params) as response:
            response.raise_for_status()
            datasets: list[dict[str, Any]] = response.json()
            return datasets

    def find_by_id(self, dataset_id: str) -> dict[str, Any]:
        """Find a dataset by id.

        ??? info "Endpoint"
            [GET /v1/datasets/{id}](https://api.prod.generalsystem.com/docs/api#/Dataset%20Management%20(v1)/get_v1_datasets__id_)

        Parameters
        ----------
        dataset_id:
            id of the dataset to retrieve.

        Returns
        -------
        dataset:
            if the Dataset ID exists returns information about the dataset.

        Raises
        ------
        DFIResponseError
            if the Dataset ID cannot be found.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)
        dfi.datasets.find_by_id(dataset_id="1234")
        ```
        """
        with self.conn.api_get(f"v1/datasets/{dataset_id}", stream=False) as response:
            response.raise_for_status()
            dataset: dict[str, Any] = response.json()
            return dataset

    def update(self, dataset_id: str, dataset: dict[str, Any]) -> dict[str, Any]:
        """Update a dataset by id.

        ??? info "Endpoint"
            [PATCH /v1/datasets/{id}](https://api.prod.generalsystem.com/docs/api#/Dataset%20Management%20(v1)/patch_v1_datasets__id_)

        ??? tip "Admin Request"
            You need to be an admin for this request.

        Parameters
        ----------
        dataset_id:
            id of the dataset to update.
        dataset:
            new dataset definition.

        Returns
        -------
        dataset:
            the updated dataset info.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)

        dataset_id = "1234"

        dataset_update = {
            "description": "a test dataset",
            "dataDescription": {
                "minDatetime": "2020-01-01T00:00:00.000Z",
                "maxDatetime": "2021-01-01T00:00:00.000Z",
            },
        }

        dfi.datasets.update(dataset_id, dataset_update)
        ```

        """
        with self.conn.api_patch(
            f"v1/datasets/{dataset_id}", stream=False, json=dataset
        ) as response:
            response.raise_for_status()
            dataset_info: dict[str, Any] = response.json()
            return dataset_info

    def delete(self, dataset_id: str) -> None:
        """Delete a dataset by id.

        ??? info "Endpoint"
            [DELETE /v1/datasets/{id}](https://api.prod.generalsystem.com/docs/api#/Dataset%20Management%20(v1)/delete_v1_datasets__id_)

        ??? tip "Admin Request"
            You need to be an admin for this request.

        Parameters
        ----------
        dataset_id:
            id of the dataset to delete.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)

        dataset_id = "1234"
        dfi.datasets.delete(dataset_id)
        ```

        """
        with self.conn.api_delete(
            f"v1/datasets/{dataset_id}", stream=False
        ) as response:
            response.raise_for_status()
            return None

    def get_permissions(self, dataset_id: str) -> list[dict[str, Any]]:
        """Get list of permissions for this dataset.

        ??? info "Endpoint"
            [GET /v1/datasets/{id}/permissions](https://api.prod.generalsystem.com/docs/api#/Dataset%20Management%20(v1)/get_v1_datasets__id__permissions)

        Parameters
        ----------
        dataset_id:
            id of the dataset to retrieve permissions from.

        Returns
        -------
        permissions:
            list of permissions for dataset.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)
        dfi.datasets.get_permissions(dataset_id="1234")
        ```

        """
        with self.conn.api_get(
            f"v1/datasets/{dataset_id}/permissions", stream=False
        ) as response:
            response.raise_for_status()
            permissions: list[dict[str, Any]] = response.json()
            return permissions

    def add_permissions(
        self, dataset_id: str, permissions: list[dict[str, Any]]
    ) -> list[dict[str, Any]]:
        """Add new permissions to a dataset.

        ??? info "Endpoint"
            [POST /v1/datasets/{id}/permissions](https://api.prod.generalsystem.com/docs/api#/Dataset%20Management%20(v1)/post_v1_datasets__id__permissions)

        ??? tip "Admin Request"
            You need to be an admin for this request.

        Parameters
        ----------
        dataset_id:
            id of the dataset to add permissions to.
        permissions:
            list of permissions to be added.

        Returns
        -------
        permissions:
            list of added permissions for the dataset.

        Examples
        --------
        ### I. Update permissions so anyone can read the dataset
        ```python
        from dfi import Client

        dataset_id = "1234"
        permissions = [{"type": "reader", "scope": "all"}]
        dfi = Client(token, url)
        updated_permissions = dfi.datasets.add_permissions(
            dataset_id=dataset_id, permissions=permissions
        )
        ```

        ### II. Update permissions specific user has write permissions
        ```python
        from dfi import Client

        dataset_id = "1234"
        permissions = [
            {
                "type": "writer",
                "scope": "identity",
                "identityId": "user-123",
            }
        ]
        dfi = Client(token, url)
        updated_permissions = dfi.datasets.add_permissions(
            dataset_id=dataset_id, permissions=permissions
        )
        ```

        """
        with self.conn.api_post(
            f"v1/datasets/{dataset_id}/permissions", stream=False, json=permissions
        ) as response:
            response.raise_for_status()
            updated_permissions: list[dict[str, Any]] = response.json()
            return updated_permissions

    def delete_permissions(
        self, dataset_id: str, permissions: list[dict[str, Any]]
    ) -> list[dict[str, Any]]:
        """Remove permissions from a dataset.  Given permission must match exactly to be removed.

        ??? info "Endpoint"
            [DELETE /v1/datasets/{id}/permissions](https://api.prod.generalsystem.com/docs/api#/Dataset%20Management%20(v1)/delete_v1_datasets__id__permissions)

        ??? tip "Admin Request"
            You need to be an admin for this request.

        Parameters
        ----------
        dataset_id:
            id of the dataset to remove permissions from.
        permissions:
            permissions to be deleted from the dataset.

        Returns
        -------
        permissions:
            list of deleted permissions from the dataset.

        Examples
        --------
        ```python
        from dfi import Client

        dataset_id = "1234"
        permissions = [
            {
                "type": "reader",
                "scope": "identity",
                "identityId": "123",
            },
            {"type": "reader", "scope": "all"},
        ]
        dfi = Client(token, url)
        dfi.datasets.delete_permissions(
            dataset_id=dataset_id, permissions=permissions
        )
        ```

        """
        with self.conn.api_delete(
            f"v1/datasets/{dataset_id}/permissions", stream=False, json=permissions
        ) as response:
            response.raise_for_status()
            removed_permissions: list[dict[str, Any]] = response.json()
            return removed_permissions

    def get_my_permissions(self, dataset_id: str) -> dict:
        """Get list of the current identity's permissions on a dataset.

        ??? info "Endpoint"
            [GET /v1/datasets/{id}/permissions/me](https://api.prod.generalsystem.com/docs/api#/Dataset%20Management%20(v1)/get_v1_datasets__id__permissions_me)

        Parameters
        ----------
        dataset_id:
            id of the dataset to retrieve user permissions from.

        Returns
        -------
        dataset permissions:
            the user's permissions for the dataset.

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)
        dfi.datasets.get_my_permissions(dataset_id="1234")
        ```

        """
        with self.conn.api_get(
            f"v1/datasets/{dataset_id}/permissions/me", stream=False
        ) as response:
            response.raise_for_status()
            permissions: dict[str, Any] = response.json()
            return permissions

    def get_schema(
        self, dataset_id: str, schema_type: str = "full", media_type: str = "json"
    ) -> dict[str, Any] | pa.Schema:
        """Retrieve a copy of the schema for this dataset.

        ??? info "Endpoint"
            [GET /v1/datasets/{id}/schema](https://api.prod.generalsystem.com/docs/api#/Dataset%20Management%20(v1)/get_v1_datasets__id__schema)

        Parameters
        ----------
        dataset_id:
            id of the dataset to retrieve schema from.
        schema_type:
            set which type of Feather schema to return.  Defaults to "full".

            - `full` - Return a full Feather schema with all fields
            - `core` - Return a Feather schema with only the core fields
            - `withMetadataId` - return a Feather schema with the core fields + metadataId field.
            - `withFilterFields` - return a Feather schema with the core fields + filter fields.
        media_type: (`json` or `feather`) defines the file format for the returned schema, either `"json"`
            or `"feather"`.  Defaults to `"json"`.

        Returns
        -------
        schema:
            the dataset's schema.

        Examples
        --------
        ### Retrieve dataset schema as JSON
        ```python
        from dfi import Client

        dfi = Client(token, url)
        schema = dfi.datasets.get_schema(
            dataset_id="<dataset id>",
            schema_type="full",
            media_type="json",
        )
        ```

        ### Retrieve dataset schema as Feather bytes
        ```python
        from dfi import Client

        dfi = Client(token, url)
        schema = dfi.datasets.get_schema(
            dataset_id="<dataset id>",
            schema_type="full",
            media_type="feather",
        )
        ```

        """
        schema_types = {
            "full": "full",
            "core": "core",
            "with_metadata_id": "withMetadataId",
            "with_filter_fields": "withFilterFields",
        }
        params = {"type": schema_types[schema_type]}

        headers = {
            "Authorization": f"Bearer {self.conn.api_token}",
            "Accept": (
                "application/feather" if media_type == "feather" else "application/json"
            ),
        }
        url = f"{self.conn.base_url}/v1/datasets/{dataset_id}/schema"

        with requests.get(
            url,
            headers=headers,
            stream=False,
            params=params,
            timeout=self.conn.query_timeout,
        ) as response:
            response.raise_for_status()

            if media_type == "feather":
                buffer = memoryview(response.content)
                with pa.input_stream(buffer) as stream:
                    schema = stream.read_buffer()
                    return feather.read_table(schema).schema
            return response.json()

    def add_enums(
        self, dataset_id: str, metadata_enums: dict[str, Any]
    ) -> dict[str, Any]:
        """Add new values to an enum field. Any new values added to fields here are merged into the existing values.

        ??? info "Endpoint"
            [POST /v1/datasets/{id}/schema/values](https://api.prod.generalsystem.com/docs/api#/Dataset%20Management%20(v1)/post_v1_datasets__id__schema_values)

        Parameters
        ----------
        dataset_id:
            id of the dataset to add enum values to.
        metadata_enums:
            a dictionary of metadata fields with enums values to be added to the dataset schema.

        Returns
        -------
        schema:
            the newly updated metadata schema

        Examples
        --------
        ```python
        from dfi import Client

        dfi = Client(token, url)

        metadata_enums = {
            "plantCultivar": {
                "type": "enum",
                "values": ["kale", "kohlrabi", "mustard"],
                "nullable": True,
            },
        }
        dfi.datasets.add_enums(
            dataset_id="1234", metadata_enums=metadata_enums
        )
        ```
        """
        with self.conn.api_post(
            f"v1/datasets/{dataset_id}/schema/values",
            stream=False,
            json=metadata_enums,
        ) as response:
            response.raise_for_status()
            schema: dict[str, Any] = response.json()
            return schema
